'''
Contains generic widgets customisation for matplotlib qt backend
For example specific types of sliders (float, min max)
'''

import sys
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_qtagg import FigureCanvas
from matplotlib.backends.backend_qtagg import NavigationToolbar2QT as NavigationToolbar

# from PySide6 import QtWidgets, QtCore
from matplotlib.backends.qt_compat import QtWidgets, QtCore
import matplotlib

import scabbard as scb

# from scabbard.steenbok.gui_colormap_picker import str2cmap, ColorMapWidget

# Ensure using the Qt6Agg backend with PySide6
matplotlib.use('QtAgg')






class MapWidget(QtWidgets.QWidget):
	'''
	MapWidget, plots a topographic map from an environment
	'''

	
	def __init__(self, env):
		super(MapWidget, self).__init__()

		# this widget is made to be connected to an environment, for updates
		self.env = env

		# grid layout is the easiest layout to organise everything by row and col
		self.grid_layout = QtWidgets.QGridLayout(self)

		# Create two matplotlib figures and set their canvases
		self.figure, self.ax1 = plt.subplots()
		self.canvas = FigureCanvas(self.figure)


		# Add canvases and toolbars to the layout
		self.grid_layout.addWidget(NavigationToolbar(self.canvas, self), 0, 0)  # Row 1, Column 0
		self.grid_layout.addWidget(self.canvas, 1, 0)  # Row 0, Column 0

		# Initial plots
		self.imTopo = self.ax1.imshow(self.env.grid.Z2D, cmap='gist_earth')
		self.imHS = self.ax1.imshow(self.env.grid.hillshade, cmap='gray', alpha = 0.6)

		self.drapePlot = None

		self.ax1.set_xlabel("X (m)")
		self.ax1.set_ylabel("Y (m)")

		# Controls layout for the first plot
		self.controls_layout = QtWidgets.QHBoxLayout()
		self.grid_layout.addLayout(self.controls_layout, 2, 0)  # Row 2, Column 0



		# Draw the canvases
		self.canvas.draw()
		# self.canvas2.draw()




def map_widget_from_fname(fname):
	'''
	Loads a map widget from a file
	'''
	
	env = scb.env_from_DEM(fname)

	return MapWidget(env)

def map_widget_from_fname_for_graphflood(fname):
	'''
	Loads a map widget from a file
	'''
	
	env = scb.env_from_DEM(fname)
	mapw = MapWidget(env)
	mapw.imHS.set_alpha(1.)
	mapw.drapePlot = mapw.ax1.imshow(np.zeros_like(env.grid.Z2D), cmap='Blues', alpha = 0.6, vmin = 0, vmax = 0.8)
	plt.colorbar(mapw.drapePlot, label = 'Flow depth (m)')

	return mapw


# Plotting wizard
import click
import scabbard as scb
import matplotlib.pyplot as plt
import dagger as dag
import numpy as np


# ANSI escape sequences for colors
RESET = "\x1b[0m"
RED = "\x1b[31m"
GREEN = "\x1b[32m"
YELLOW = "\x1b[33m"
BLUE = "\x1b[34m"
MAGENTA = "\x1b[35m"
CYAN = "\x1b[36m"


@click.command()
@click.argument('fname', type = str)
def simplemapwizard(fname):
	plt.ioff()
	dem = scb.raster2RGrid(fname)
	atlas = scb.Dplot.basemap(dem)
	plt.show()
	input("press Enter to continue")










	


@click.command()
@click.option('-c', '--courant', 'courant',  type = float, default = 1e-3)
@click.option('-d', '--dt', 'dt',  type = float, default = None)
@click.option('-P', '--precipitations', 'precipitations',  type = float, default = 30)
@click.option('-m', '--manning', 'manning',  type = float, default = 0.033)
@click.option('-S', '--SFD', 'SFD', type = bool, default=False)
@click.option('-U', '--update_step', 'nupdate', type = int, default=10)
@click.option('-exp', '--experimental', 'experimental', type = bool, default=False, is_flag = True)
# @click.option('-h', '--help', 'help', type = bool, is_flag = True)
@click.argument('fname', type = str)
def graphflood_basic(fname,courant,dt,precipitations,manning,SFD,nupdate, experimental):
	print("EXPERIMENTAL IS ", experimental)

	P = precipitations /1e3/3600
	if(dt is not None and courant == 1e-3):
		print(RED + "WARNING, dt set to constant value AND courant, is that normal?\n\n" + RESET)

	print("+=+=+=+=+=+=+=+=+=+=+=+=+")
	print("+=+=+=GRAPHFLOOD+=+=+=+=+")
	print("+=+=+=+=+=+=+=+=+=+=+=+=+\n\n")

	print("Precipitation rates = ", P, " m/s (",precipitations," mm/h)")


	mod = scb.ModelHelper()
	mod.init_dem_model(fname, sea_level = 0., P = precipitations)

	# Number of nodes in the Y direction
	ny = mod.grid.ny
	nx = mod.grid.nx


	mod.courant = False if(dt is not None) else True
	mod.stationary = True 
	# mod.gf.flood.enable_Qwout_recording()

	# manning friction:
	mod.mannings = manning

	# Single flow solver?
	mod.SFD = SFD;


	mod.dt = dt if dt is not None else 1e-3
	mod.min_courant_dt = 1e-6
	mod.courant_number = courant

	ph = scb.PlotHelper(mod)
	ph.init_hw_plot(use_extent = False)

	update_fig = nupdate
	i = 0
	j = 0
	while True:
		i+=1
		mod.run() if(experimental == False) else mod.gf.flood.run_hydro_only()
		if(i % update_fig > 0):
			continue
		hw = mod.gf.flood.get_hw().reshape(mod.gf.grid.rshp)
		print("Running step", i)
		ph.update()



@click.command()
@click.argument('fname', type = str)
@click.option('-p', '--precipitations', 'precipitations', type = float, default=30)
@click.option('-d', '--dt', 'dt', type = float, default=1e-3)
@click.option('-v', '--visu', 'visu', type = bool, default=False, is_flag = True)
@click.argument('output', type = str, default='flow_depth.tif')
def GPUgraphflood(fname, precipitations, dt, visu, output):

	from scabbard.riverdale.rd_params import param_from_dem
	from scabbard.riverdale.rd_env import create_from_params
	import scabbard.riverdale.rd_morphodynamics as rdmo
	import scabbard.riverdale.rd_LM as rdlm
	import taichi as ti
	import time 

	ti.init(ti.vulkan)

	param = param_from_dem(fname)
	param.dt_hydro = dt
	param.P = precipitations * 1e-3/3600
	param.set_boundary_slope(0,mode = 'elevation')
	rd = create_from_params(param)

	rdlm.priority_flood(rd,Zw = False)

	if(visu):
		fig,ax = plt.subplots()
		imhw = ax.imshow(np.zeros( (param._ny, param._nx) ), cmap = "magma", vmin = 0.8, vmax = 1.2)
		plt.colorbar(imhw, label = "Convergence (1 is perfect)")
		fig.show()
		fig.canvas.draw_idle()
		fig.canvas.start_event_loop(0.01)

	NN = 1000
	for i in range(round(1e6)):
		st = time.time()
		# rdlm.priority_flood(rd)
		rd.run_hydro(NN)
		if(visu):
			data = rd.QwC.to_numpy()/rd.QwA.to_numpy()
			data[~np.isfinite(data)] = 1
			imhw.set_data(data)
			# imhw.set_clim(np.percentile(data,20),np.percentile(data,99.5))

			fig.canvas.draw_idle()
			fig.canvas.start_event_loop(0.01)

		conv = rd.convergence_ratio
		took = time.time() - st

		print('It:',i*NN+1,':',conv, " took ", took, "seconds")
		if(conv > 0.90):
			break

	np.savez_compressed(f'results_{precipitations}.npz', {'hw':rd.hw.to_numpy()})

	fig,ax = plt.subplots()
	cb=ax.imshow(rd.hw.to_numpy(), cmap = 'Blues')
	plt.colorbar(cb, label = 'Flow depth (m)')
	plt.show()







@click.command()
@click.argument('fname', type = str)
def _debug_1(fname):
	plt.ioff()
	dem = scb.raster2RGrid(fname)
	atlas = scb.Dplot.basemap(dem)
	atlas.fig.show()
	plt.pause(0.01)
	while(True):
		plt.pause(1)
		dem.add_random_noise(-10,10)
		atlas.update()
	input("press Enter to continue")



@click.command()
def haguid():
	from scabbard.haguid import launch_haGUId

	launch_haGUId()

@click.command()
def run_nice_haguid():
	import runpy
	import os
	import sys

	# Determine the path to the script directory
	script_dir = os.path.dirname(os.path.abspath(__file__))
	# script_path = os.path.join(script_dir, 'my_script.py')
	# sub.run(f'python {os.path.join(script_dir, "nice_haguid.py")}', check = True, shell = True)
	runpy.run_path(os.path.join(script_dir, "nice_haguid.py"), run_name="__main__")

@click.command()
@click.argument('fname', type = str)
def visu2Dnpy(fname):

	import sys
	import numpy as np
	import matplotlib.pyplot as plt
	from matplotlib.backends.backend_qtagg import FigureCanvas
	from matplotlib.backends.backend_qtagg import NavigationToolbar2QT as NavigationToolbar

	# from PySide6 import QtWidgets, QtCore
	from matplotlib.backends.qt_compat import QtWidgets, QtCore
	import matplotlib
	from matplotlib.lines import Line2D

	import scabbard.phineas_helper as phelp

	# Ensure using the Qt6Agg backend with PySide6
	matplotlib.use('QtAgg')

	class MatplotlibWidget(QtWidgets.QWidget):
	
		def __init__(self, data):
			super(MatplotlibWidget, self).__init__()

			self.data = data

			self.cid = None
			self.cid2 = None

			self.initUI()
			self.setWindowTitle("Scabbard numpy 2D array explorer")
			 # Get the primary screen's geometry
			screen_geometry = QtWidgets.QApplication.primaryScreen().geometry()
			width, height = screen_geometry.width() * 0.8, screen_geometry.height() * 0.8

			# Calculate the position to center the window, then move it slightly down
			x = screen_geometry.width() * 0.1
			y = (screen_geometry.height() - height) / 2 + screen_geometry.height() * 0.05

			# Set the window geometry
			self.setGeometry(int(x), int(y), int(width), int(height))

		def initUI(self):
			# Main layout
			grid_layout = QtWidgets.QGridLayout(self)

			# Create two matplotlib figures and set their canvases
			self.figure1, self.ax1 = plt.subplots()
			self.canvas1 = FigureCanvas(self.figure1)
			self.figure2, self.ax2 = plt.subplots()
			self.canvas2 = FigureCanvas(self.figure2)

			# Add canvases and toolbars to the layout
			grid_layout.addWidget(self.canvas1, 1, 0)  # Row 0, Column 0
			grid_layout.addWidget(NavigationToolbar(self.canvas1, self), 0, 0)  # Row 1, Column 0
			grid_layout.addWidget(self.canvas2, 1, 1)  # Row 0, Column 1
			grid_layout.addWidget(NavigationToolbar(self.canvas2, self), 0, 1)  # Row 1, Column 1

			# Initial plots
			self.imshow1 = self.ax1.imshow(self.data, cmap='magma')
			self.colorbar1 = self.figure1.colorbar(self.imshow1, ax=self.ax1)
			self.plot1 = Line2D([0,self.data.shape[1]], [0, 0], color='r', linestyle='--')
			self.ax1.add_line(self.plot1)
			self.plot2 = self.ax2.plot(self.data[0,:], color = 'k', lw = 4)

			# Controls layout for the first plot
			controls_layout1 = QtWidgets.QHBoxLayout()
			grid_layout.addLayout(controls_layout1, 2, 0)  # Row 2, Column 0

			


			# Colormap selection for the first plot
			self.colormapComboBox1 = QtWidgets.QComboBox()
			self.colormapComboBox1.addItems(['magma', 'viridis', 'cividis', 'Blues', 'Reds', 'RdBu_r'])
			self.colormapComboBox1.currentIndexChanged.connect(lambda: self.update_plot('1'))
			controls_layout1.addWidget(self.colormapComboBox1)

			#RangeSelection
			self.rangeSlider1 = phelp.RangeSlider(self.data.min(), self.data.max(), ID = '1')
			# grid_layout.addWidget(self.rangeSlider1,3,0)
			self.rangeSlider1.rangeChanged.connect(self._update_crange)
			controls_layout1.addWidget(self.rangeSlider1)



			# Controls layout for the second plot
			controls_layout2 = QtWidgets.QHBoxLayout()
			grid_layout.addLayout(controls_layout2, 2, 1)  # Row 2, Column 1

			# Colormap selection for the second plot
			# self.colormapComboBox2 = QtWidgets.QComboBox()
			# self.colormapComboBox2.addItems(['magma', 'viridis', 'cividis', 'Blues', 'Reds', 'RdBu_r'])
			# self.colormapComboBox2.currentIndexChanged.connect(lambda: self.update_plot('2'))
			# controls_layout2.addWidget(self.colormapComboBox2)

			self.button = QtWidgets.QPushButton("Cross section",self)
			self.button.clicked.connect(self.switch_coordinate_picking)
			controls_layout2.addWidget(self.button)

			# Draw the canvases
			self.canvas1.draw()
			self.canvas2.draw()

		def update_plot(self, plt_ID):
			cmap1 = self.colormapComboBox1.currentText()
			# cmap2 = self.colormapComboBox2.currentText()

			if plt_ID == '1':
				self.imshow1.set_cmap(cmap1)


			# Redraw the canvases
			self.canvas1.draw()

		def _update_crange(self, vmin,vmax, plt_ID):
			if plt_ID == '1':
				self.imshow1.set_clim(vmin,vmax)


			# Redraw the canvases
			self.canvas1.draw()

		def switch_coordinate_picking(self):
			# Disconnect existing connections if any
			if self.cid:
				self.canvas1.mpl_disconnect(self.cid)
				self.cid = None
			else:
				# Connect the mouse click event to the on_motion function
				self.cid = self.canvas1.mpl_connect('motion_notify_event', self.on_motion)

		def plot_CS_from_coord(self, tx, ty):
			row = round(ty) #will not be true for maps, is just for test here
			self.plot2[0].set_ydata(self.data[row,:])
			self.ax2.relim()
			self.ax2.autoscale_view(True, True, True)
			self.plot1.set_ydata([row,row])
			self.canvas1.draw()
			self.canvas2.draw()
			# self.switch_coordinate_picking()


		def on_motion(self, event):
			# Call the function with the coordinates
			if event.xdata is not None and event.ydata is not None:
				self.cid2 = self.canvas1.mpl_connect('button_press_event', self.on_click)
				self.plot_CS_from_coord(event.xdata, event.ydata)

		def on_click(self, event):
			# Call the function with the coordinates
			if event.xdata is not None and event.ydata is not None:
				self.switch_coordinate_picking()
				if self.cid2:
					self.canvas1.mpl_disconnect(self.cid2)
					self.cid2 = None



	app = QtWidgets.QApplication(sys.argv)
	data = np.load(fname)
	if(len(data.shape) != 2):
		raise ValueError("Needs to be 2D numpy array")
	main = MatplotlibWidget(data)
	main.show()
	sys.exit(app.exec_())



'''
Sets of function to compute hydrodynamics with RiverDale

B.G. - 29/04/2024
'''

import taichi as ti
import numpy as np
from enum import Enum
import scabbard.utils as scaut 
from scabbard.riverdale.rd_grid import GRID
import scabbard.riverdale.rd_grid as gridfuncs
import scabbard.riverdale.rd_drainage_area as rda
import scabbard.riverdale.rd_helper_surfw as hsw
import scabbard.riverdale.rd_LM as rlm



@ti.kernel
def _compute_hydraulic_gradient(Sw:ti.template(),Z:ti.template(), hw:ti.template(), BCs:ti.template()):
	'''
	Taichi kernel to compute the hydraulic gradient (as topographic analysis)
	
	Arguments:
		- Sw: field of hydrualic gradient to be edited in place
		- Z: field of topographic surface
		- hw: field of flow depth
		- BCs: field of boundary conditions
	
	Returns:
		- nothing, but calculates Sw

	Authors:
		- B.G. (06/2024)
	'''

	for i,j in Z:
		if(gridfuncs.is_active(i,j,BCs) and gridfuncs.can_receive(i,j,BCs)):
			Sw[i,j] = hsw.hydraulic_gradient_value(Z, hw, BCs, i, j) 
		else:
			Sw[i,j] = 0.

def compute_hydraulic_gradient(rd, fill_with_PF = False):
	'''
	Computes hydraulic gradient as the norm of Z + hw:
	
	This function is optimised for analysis, not for internal calculation

	Arguments:
		- rd: the riverdale object, initialised
		- fill_with_PF: fill the local minimas with water using an adapted priority flood (Barnes, 2014) if activated
	Returns:
		- a numpy array of hydraulic gradient
	Authors:
		- B.G (last modifications: 06/2024)

	'''

	if(fill_with_PF):
		rlm.priority_flood(rd)

	output, = rd.query_temporary_fields(1,dtype = ti.f32)
	_compute_hydraulic_gradient(output, rd.Z,rd.hw,rd.BCs)
	return output.to_numpy()

@ti.kernel
def _compute_shear_stress(shear_stress:ti.template(),Z:ti.template(), hw:ti.template(), BCs:ti.template(), rho:ti.f32, g:ti.f32):
	'''
	Taichi kernel to compute the shear stress (as topographic analysis)
	
	Arguments:
		- shear_stress: field of shear stress to be edited in place
		- Z: field of topographic surface
		- hw: field of flow depth
		- BCs: field of boundary conditions
		- rho: density of water kg/m^3 (should be 1000 in "normal" cases)
		- g: gravitational acceleration in m/s^2 (9.8 in "nornal" cases)
	
	Returns:
		- nothing, but calculates the shear stress

	Authors:
		- B.G. (06/2024)
	'''



	for i,j in Z:
		if(gridfuncs.is_active(i,j,BCs) and gridfuncs.can_receive(i,j,BCs)):
			shear_stress[i,j] = hsw.hydraulic_gradient_value(Z, hw, BCs, i, j) * ti.math.max(hw[i,j],0.) * rho * g
		else:
			shear_stress[i,j] = 0.

def compute_shear_stress(rd, fill_with_PF = False):
	'''
	Computes the shear stress for every nodes with water for a given riverdale environment:
	τ = ρ g h_w S_w
	with ρ the water density, g the gravitational acceleration, h_w the flow depth and S_w the hydraulic slope

	THis function is optimised for analysis, not for internal calculation

	Arguments:
		- rd: the riverdale object, initialised
		- fill_with_PF: fill the local minimas with water using an adapted priority flood (Barnes, 2014) if activated
	Returns:
		- a numpy array of shear stress
	Authors:
		- B.G (last modifications: 06/2024)

	'''

	if(fill_with_PF):
		rlm.priority_flood(rd)

	output, = rd.query_temporary_fields(1,dtype = ti.f32)
	_compute_shear_stress(output, rd.Z,rd.hw,rd.BCs, 1000., 9.81)
	return output.to_numpy()

def compute_flow_velocity(rd, use_Qwin = True):
	'''
	Computes the flow velocity from a RiverDale object
	'''
	u = rd.QwA.to_numpy() if use_Qwin else rd.QwC.to_numpy()
	hw = rd.hw.to_numpy()
	mask0 = hw<=1e-6
	u[mask0] = 0
	u[~mask0] /= hw[~mask0]*GRID.dx

	return u


def compute_discharge_per_unit_width(rd, use_Qwin = True):
	'''
	Computes the discharge per unit width from a RiverDale object
	'''
	qw = rd.QwA.to_numpy() if use_Qwin else rd.QwC.to_numpy()
	qw /= GRID.dx
	return qw


def compute_effective_drainage_area(rd, use_Qwin = True, custom_r = None):
	'''
	Computes the effective drainage area a_r (s) as of described in e.g. Bernard et al. (2023) 

	Arguments:
		- rd: the initialised riverdale's object
		- use_Qwin: use Qwin if True, else Qwout
		- custom_r: effective drainage area is equal to Discharge per unit width divided by average precipitation rates, if this parameters is not None it replaces the one calculated from rd.param
	'''

	if(rd.param.need_input_Qw and custom_r is None):
		raise NotImplementedError("It seems that your Riverdale's Object has input points for Discharge. I cannot calculate the effective drainage area without a custom_r parameter representing the average precipitation rate.")

	if(rd.param.precipitations_are_2D):
		raise NotImplementedError("WIP, you have 2D precipitations input and I need ot implement a function to compute average upstream precipiation rate for each node.")

	if(custom_r is None):
		custom_r = rd.param.precipitations

	qw = compute_discharge_per_unit_width(rd, use_Qwin)

	return qw/custom_r


def compute_effective_width(rd, use_Qwin = True):
	'''
	EXPERIMENTAL

	Will probably evolve quite a lot in the coming weeks so I'll comment at that moment
	'''

	if(rd.param.precipitations_are_2D):
		prec = rd.param.precipitations
	else:
		prec = np.full(rd.param.reshp, rd.param.precipitations)

	QWD4 = compute_drainage_area_D4(rd, fill = True, N = 'auto', random_rec = True, Precipitations = prec)
	qw   = compute_discharge_per_unit_width(rd, use_Qwin)

	mask0 = qw <= 0

	Weff = np.zeros_like(qw)

	Weff[~mask0] = QWD4[~mask0]/qw[~mask0]

	return Weff




	





















# End of file


'''
this script defines the geometry features of given elements
They are made to be accessed from the top level (end users)


B.G.
'''

import numpy as np
from abc import ABC, abstractmethod


class BaseGeometry(ABC):
	"""
	Base abstract class for the geometry objects
	Defines all the functions that have to be defined for each geometry type, 
	even if only to throw an error if not possible (e.g. row and cols for a irregular grid)

	Authors:
		- B.G. (last modifications: 08/2024)


	"""

	def __init__(self):
		self._crs = None

	@property
	@abstractmethod
	def N(self):
		'''
		return the number of nodes in the element

		Authors:
		- B.G. (last modifications: 08/2024)
		'''
		pass

	@property
	def nxy(self):
		'''
		Alias for the number of nodes

		Authors:
		- B.G. (last modifications: 08/2024)
		'''
		return self.N

	@property
	@abstractmethod
	def dx(self):
		'''
		return the spatial step

		Authors:
		- B.G. (last modifications: 08/2024)
		'''
		pass

	@property
	@abstractmethod
	def nx(self):
		'''
		return the number of nodes in the x directions

		Authors:
		- B.G. (last modifications: 08/2024)
		'''
		pass

	@property
	def ncolumns(self):
		return self.nx


	@property
	@abstractmethod
	def ny(self):
		'''
		return the number of nodes in the y directions

		Authors:
		- B.G. (last modifications: 08/2024)
		'''
		pass

	@property
	def nrows(self):
		'''
		Alias for ny

		Authors:
		- B.G. (last modifications: 08/2024)
		'''
		return self.ny


	@property
	@abstractmethod
	def xmin(self):
		'''
		return the number of nodes in the y directions

		Authors:
		- B.G. (last modifications: 08/2024)
		'''
		pass

	@property
	def Xmin(self):
		return self.xmin

	@property
	@abstractmethod
	def xmax(self):
		'''
		return the number of nodes in the y directions

		Authors:
		- B.G. (last modifications: 08/2024)
		'''
		pass

	@property
	def Xmax(self):
		return self.xmax

	@property
	@abstractmethod
	def ymin(self):
		'''
		return the number of nodes in the y directions

		Authors:
		- B.G. (last modifications: 08/2024)
		'''
		pass

	@property
	def Ymin(self):
		return self.ymin

	@property
	@abstractmethod
	def ymax(self):
		'''
		return the number of nodes in the y directions

		Authors:
		- B.G. (last modifications: 08/2024)
		'''
		pass

	@property
	def Ymax(self):
		return self.ymax


	@property
	@abstractmethod
	def shape(self):
		pass

	@property
	def crs(self):
		return crs

	@abstractmethod
	def row_col_to_flatID(self, row, col):
		'''
		Take row col (single or array) and returns the flat index for regular datas

		Authors:
		- B.G. (last modifications: 08/2024)
		'''
		pass

	@abstractmethod
	def flatID_to_row_col(self, flatID):
		'''
		Take row col (single or array) and returns the flat index for regular datas

		Authors:
		- B.G. (last modifications: 08/2024)
		'''
		pass


	@abstractmethod
	def row_col_to_X_Y(self, row, col):
		'''
		Converts row col (for regular grids) to X Y coordinates (real world)

		Authors:
		- B.G. (last modifications: 08/2024)
		'''

		pass



	@abstractmethod
	def X_Y_to_row_col(self, X, Y):
		'''
		Converts  X Y coordinates (real world) to row col (for regular grids)

		Authors:
		- B.G. (last modifications: 08/2024)
		'''

		pass

	@abstractmethod
	def flatID_to_X_Y(self, flatID):
		'''
		Converts  flat ID to XY coordinates

		Authors:
		- B.G. (last modifications: 08/2024)
		'''

		pass

	@abstractmethod
	def X_Y_to_flatID(self, X, Y):
		'''
		Converts  X Y coordinates (real world) to flat ID (for regular grids)

		Authors:
		- B.G. (last modifications: 08/2024)
		'''

		pass


	@property
	def extent(self):
		'''
		Matplotlib friendly extent argument for imshow or more generally any bounding box

		Authors:
		- B.G. (last modifications: 08/2024)

		'''
		return [self.xmin,self.xmax, self.ymax, self.ymin]





'''
High level grpahflood object
'''
import dagger as dag
import scabbard as scb
import numpy as np
import matplotlib.pyplot as plt
import cmcrameri.cm as cm




class GraphFlood(object):

	"""
		docstring for GraphFlood
	"""

	def __init__(self, 
		grid,
		verbose = False,
		convergence_tracking = True,
		**kwargs
		):

		super(GraphFlood, self).__init__()
		
		# Grid object
		self.grid = grid

		# Checking and or feeding the graph
		if (self.grid.con is None and self.grid.graph is None):
			self.grid.compute_graphcon()
		elif (self.grid.con is not None and self.grid.graph is None):
			self.grid.graph = dag.graph(self.grid.con)
			self.grid.graph.compute_graph(self._Z, False, True)
		
		# Initialising the c++ graphflood object
		self.flood = dag.graphflood(self.grid.graph, self.grid.con)

		# feeding the topo
		self.flood.set_topo(self.grid._Z)

		# run all the configuration options
		self.config(**kwargs)

		self.active_figs = []
		self._callbacks = []

		self.hydro_dt = 1e-3 	

		self.verbose = verbose

		self.update_grid = True


		# time
		self.cumulative_time_hydro = 0.
		self.nit_hydro = 0
		self.cumulative_time_morpho = 0.

		# Convergence criteriae
		self.init_convergence_tracking(convergence_tracking)

	def config(self, **kwargs):
		'''
			Configures some of the general model options and inputs.
			Wrapper on the c++ object, not all options are in to keep things clear.
			Options:
				- dt or hydro_dt: hydrologic time ste
				- SFD (bool): if True set the mode to single flow, else multiple flow
				- minima (str): configure the method used for local minima resolution
					+ "reroute": reroute flow without filling the minimas
					+ "ignore": stop the flow at local minimas
					+ "fill" (default AND reccomended): fill the local minima with water

		'''

		if("SFD" in kwargs.keys()):
			if(kwargs["SFD"] == True):
				print("set flow conditions to single flow direction") if self.verbose else 0
				self.flood.enable_SFD()
			else:
				print("set flow conditions to mutliple flow direction") if self.verbose else 0
				self.flood.enable_MFD()

		if("minima" in kwargs.keys()):
			if(kwargs["minima"].lower() == "reroute"):
				self.flood.reroute_minima()
			elif(kwargs["minima"].lower() == "ignore"):
				self.flood.ignore_minima()
			else:
				self.flood.fill_minima()

		if("dt" in kwargs.keys()):
			self.flood.set_dt_hydro(kwargs["dt"])
			self.hydro_dt = kwargs["dt"]

		if("hydro_dt" in kwargs.keys()):
			self.flood.set_dt_hydro(kwargs["hydro_dt"])
			self.hydro_dt = kwargs["hydro_dt"]

		if("n_trackers" in kwargs.keys()):
			self.n_trackers = kwargs['n_trackers']

	def init_convergence_tracking(self, yes = True):

		self.convergence_trackers = yes
		self.n_trackers = 50
		self.n_pits = []
		self.dhw_monitoring = []
		self.Qwratio = []

		if(yes):
			self.flood.enable_Qwout_recording()
			self.flood.enable_dhw_recording()

	def set_precipitations(self,values = 1e-4):

		if(isinstance(values, np.ndarray)):
			if(np.prod(values.shape) == self.grid.nxy ):
				self.flood.set_water_input_by_variable_precipitation_rate(values.ravel())
			else:
				raise RuntimeError("array of precipitations needs to be of grid size or a single scalar")

		else:
			self.flood.set_water_input_by_constant_precipitation_rate(values)


	def set_input_discharge(self, nodes, values, asprec = False):
		'''
			Sets the input discharge at given locations (nodes need to be in flattened node indices)
		'''

		if(isinstance(nodes,list)):
			nodes = np.array(nodes, dtype=np.int32)
		if(isinstance(values,list)):
			values = np.array(values, dtype=np.float64)
		if(np.prod(nodes.shape) != np.prod(values.shape)):
			raise RuntimeError(f"nodes and values need to be the same size. Currently {np.prod(nodes)} vs {np.prod(values)}")

		# YOLO
		if(asprec == False):
			self.flood.set_water_input_by_entry_points(values, nodes)
		else:
			P = np.zeros(self.grid.nxy)
			P[nodes] = values
			self.flood.set_water_input_by_variable_precipitation_rate( P.ravel())


	def run_hydro(self, n_steps = 1, fig_update_step = 1,
	 force_morpho = False, run_morpho_every = 5, min_iteration_morpho = 500, run_morpho_ratio = 0.95,
	 runtime_callback = [], RAT = False, RAT_step = 1000, courant = True, **kwargs):
		'''
		'''
		# Only hydro on this version
		if(force_morpho):
			self.flood.enable_morpho()
		else:
			self.flood.disable_morpho()

		if(courant):
			self.flood.enable_courant_dt_hydro()
		else:
			self.flood.set_dt_hydro(self.hydro_dt)
		# print("DEBUG::DT is", self.hydro_dt)

		# Running loop
		for i in range(n_steps):

			# if i> min_iteration_morpho and i%run_morpho_every == 0 and force_morpho:
			totin = self.flood.get_tot_Qw_input()
			if(totin >0 and self.flood.get_tot_Qw_output() / totin > run_morpho_ratio and force_morpho):
				# print("DEGUB::MORPHO ON")
				self.flood.enable_morpho()
				if("block_uplift_rate" in kwargs):
					self.flood.block_uplift(kwargs["block_uplift_rate"])
			else:
				self.flood.disable_morpho()
			# if(i > 5000 and i%5 == 0):
			# 	input("tttt")

			self.cumulative_time_hydro += self.flood.get_dt_hydro()
			self.nit_hydro += 1

			# Running hte actual model
			self.flood.run()

			# Balance checker for debugging purposes
			balance = abs(self.flood.get_tot_Qw_input() - self.flood.get_tot_Qwin_output())
			if(balance > 1):
				print("WARNING::Qw-in imbalance -> " + str(self.flood.get_tot_Qw_input()) + " got in and " + str(self.flood.get_tot_Qwin_output()) + " got out. Unbalance is: " + str(balance))

			if(i%500 == 0):
				print(i)

			if(RAT):
				if(i>0 and i%RAT_step == 0):
					newdt = input("new_dt: ")
					if(newdt == ''):
						newdt = self.hydro_dt
					else:
						self.hydro_dt = float(newdt)

					self.flood.set_dt_hydro(self.hydro_dt)


			for rtcb in runtime_callback:
				if(isinstance(rtcb,list)):
					rtcb[0](*rtcb[1])
				else:
					rtcb()

			# Monitoring the water convergence now:
			if(self.convergence_trackers):
				if(len(self.n_pits) == self.n_trackers):
					self.n_pits.pop(0)
					self.dhw_monitoring.pop(0)
					self.Qwratio.pop(0)

				self.n_pits.append(self.grid.graph.get_n_pits())
				arr = self.flood.get_dhw_recording()
				self.dhw_monitoring.append([np.percentile(arr,5), np.percentile(arr,25), np.median(arr), np.percentile(arr,75), np.percentile(arr,95)])
				arr = self.flood.get_Qwout_recording()
				mask = arr == 0
				arr = self.flood.get_Qwin()/arr
				arr[mask] = 1
				self.Qwratio.append([np.percentile(arr,5), np.percentile(arr,25), np.median(arr), np.percentile(arr,75), np.percentile(arr,95)])

				if(self.verbose):
					print("\n###############################")
					print("###############################")
					print("Monitoring results:")
					print("N pits:",self.n_pits[-1])
					print("delta hw:",self.dhw_monitoring[-1])
					print("Qwratio:",self.Qwratio[-1])
					print("###############################")
					print("###############################\n")

			

			# updating the figures
			if(i%fig_update_step == 0):
				self.update_figs()
				if self.update_grid:
					self.grid._Z = self.flood.get_bedrock_topo()
		if self.update_grid:
			self.grid._Z = self.flood.get_bedrock_topo()

	def update_figs(self):
		for tax in self._callbacks:
			tax.update()
		for tf in self.active_figs:
			tf.canvas.draw_idle()
			tf.canvas.start_event_loop(0.001)


	def get_xmonitor_HYDRO(self):
		tx1 = max(self.nit_hydro - self.n_trackers, 0)
		tx2 = self.nit_hydro
		# print("DEBUG_XMONO::",tx1,tx2)
		return np.arange(tx1 + 1, tx2 + 1)

	def get_monitor_pits(self):
		return [self.get_xmonitor_HYDRO(), self.n_pits]

	def get_monitor_median_hw_pits(self):
		# print(self.dhw_monitoring[2], "KKKK")
		return [self.get_xmonitor_HYDRO(), np.array(self.dhw_monitoring)[:,2]]

	def get_monitor_10th_hw_pits(self):
		return [self.get_xmonitor_HYDRO(), np.array(self.dhw_monitoring)[:,0]]


	def get_monitor_90th_hw_pits(self):
		return [self.get_xmonitor_HYDRO(), np.array(self.dhw_monitoring)[:,4]]

	def debugyolo(self):
		return np.array(self.grid.graph.get_debug_mask())

	def pop_Qw_fig(self, jupyter = False, clim = None, alpha_hillshade = 0.3):

		if(jupyter == False):
			plt.ioff()

		fig,ax = plt.subplots()
		dax = scb.RGridDax(self.grid, ax, alpha_hillshade=1)
		arr = self.flood.get_Qwin()
		if(np.prod(arr.shape) == 0):
			arr = self.grid.zeros()
		else:
			arr.reshape(self.grid.rshp)

		Qwax = dax.drape_on(arr, cmap = "Blues", clim = clim, delta_zorder = 1, alpha = 1 - alpha_hillshade, callback = self.flood.get_Qwin)
		# Qwax = dax.drape_on(arr, cmap = "Blues", clim = None, delta_zorder = 1, alpha = 0.9, callback = self.debugyolo)

		self.active_figs.append(fig)
		self._callbacks.append(dax)
		self._callbacks.append(Qwax)
		fig.show()
		fig.canvas.draw_idle()
		fig.canvas.start_event_loop(0.001)

	def pop_hw_fig(self, jupyter = False, clim = None, alpha_hillshade = 0.3):

		if(jupyter == False):
			plt.ioff()

		fig,ax = plt.subplots()
		dax = scb.RGridDax(self.grid, ax, alpha_hillshade=1)
		arr = self.flood.get_Qwin()
		if(np.prod(arr.shape) == 0):
			arr = self.grid.zeros()
		else:
			arr.reshape(self.grid.rshp)

		Qwax = dax.drape_on(arr, cmap = "Blues", clim = clim, delta_zorder = 1, alpha = 1 - alpha_hillshade, callback = self.flood.get_hw)

		plt.colorbar(Qwax.im, label = "Water depth (m)")
		self.active_figs.append(fig)
		self._callbacks.append(dax)
		self._callbacks.append(Qwax)
		fig.show()
		fig.canvas.draw_idle()
		fig.canvas.start_event_loop(0.001)


	def pop_custom_fig(self, jupyter = False, clim = None, callback = None,callback_params = None, cmap = 'magma', alpha_hillshade = 0.3):

		if(callback is None):
			raise RuntimeError("Need callback for custom fig")

		if(jupyter == False):
			plt.ioff()

		fig,ax = plt.subplots()
		dax = scb.RGridDax(self.grid, ax, alpha_hillshade=1)
		arr = self.flood.get_Qwin()
		if(np.prod(arr.shape) == 0):
			arr = self.grid.zeros()
		else:
			arr.reshape(self.grid.rshp)

		Qwax = dax.drape_on(arr, cmap = cmap, clim = clim, delta_zorder = 1, alpha = 1 - alpha_hillshade, callback = callback, callback_params = callback_params)

		plt.colorbar(Qwax.im, label = "Water depth (m)")
		self.active_figs.append(fig)
		self._callbacks.append(dax)
		self._callbacks.append(Qwax)
		fig.show()
		fig.canvas.draw_idle()
		fig.canvas.start_event_loop(0.001)


	def pop_topo_fig(self, jupyter = False, clim = None, bedrock = False, alpha_hillshade =0.5):

		if(jupyter == False):
			plt.ioff()

		fig,ax = plt.subplots()
		dax = scb.RGridDax(self.grid, ax, alpha_hillshade=1)
		arr = self.flood.get_Qwin()
		if(np.prod(arr.shape) == 0):
			arr = self.grid.zeros()
		else:
			arr.reshape(self.grid.rshp)

		topo = dax.drape_on(arr, cmap = "gist_earth", clim = clim, delta_zorder = 1, alpha =1 - alpha_hillshade, callback = self.flood.get_bedrock_topo if bedrock else self.flood.get_surface_topo)

		plt.colorbar(topo.im, label = "Surface (hw + Z) (m)" if bedrock == False else "Z (m)")
		self.active_figs.append(fig)
		self._callbacks.append(dax)
		self._callbacks.append(topo)
		fig.show()
		fig.canvas.draw_idle()
		fig.canvas.start_event_loop(0.001)

	def pop_dtopo_fig(self, jupyter = False, clim = None, alpha_hillshade = 0.5):

		if(jupyter == False):
			plt.ioff()

		self.otopo = np.copy(self.grid.Z2D)

		fig,ax = plt.subplots()
		dax = scb.RGridDax(self.grid, ax, alpha_hillshade=1)
		arr = self.flood.get_Qwin()
		if(np.prod(arr.shape) == 0):
			arr = self.grid.zeros()
		else:
			arr.reshape(self.grid.rshp)

		def _pop_dtopo_fig():
			return self.otopo - self.flood.get_bedrock_topo().reshape(self.grid.rshp)

		topo = dax.drape_on(arr, cmap = cm.bam, clim = clim, delta_zorder = 1, alpha = 1 - alpha_hillshade, callback = _pop_dtopo_fig)

		plt.colorbar(topo.im, label = "$\Delta Z$")
		self.active_figs.append(fig)
		self._callbacks.append(dax)
		self._callbacks.append(topo)
		fig.show()
		fig.canvas.draw_idle()
		fig.canvas.start_event_loop(0.001)



	def pop_monitor_fig(self, jupyter = False):

		if(self.convergence_trackers == False):
			raise AttributeError("cannot pop monitor figure if convergence_trackers is not activated")

		if(jupyter == False):
			plt.ioff()

		# # gridspec inside gridspec
		# fig = plt.figure()
		# gs0 = gridspec.GridSpec(1, 2, figure=fig)

		fig,ax1 = plt.subplots()

		ax2 = ax1.twinx()

		ax1.set_xlabel('N iterations')
		ax1.set_ylabel('N internal pits')

		l1 = ax1.plot([0],[0], lw = 2, color = 'r')

		l3 = ax2.plot([0],[0], lw = 1, ls = '--', color = 'k')
		l4 = ax2.plot([0],[0], lw = 1, ls = '--', color = 'k')
		l2 = ax2.plot([0],[0], lw = 2, color = 'k')

		dax1 = scb.callbax_sline(ax1, l1, self.get_monitor_pits, axylim = None)

		dax2 = scb.callbax_sline(ax2, l2, self.get_monitor_median_hw_pits, axylim = None)
		dax3 = scb.callbax_sline(ax2, l3, self.get_monitor_10th_hw_pits, axylim = None, axylim_ignore = True)
		dax4 = scb.callbax_sline(ax2, l4, self.get_monitor_90th_hw_pits, axylim = None, axylim_ignore = True)
		


		self.active_figs.append(fig)
		self._callbacks.append(dax1)
		self._callbacks.append(dax2)
		self._callbacks.append(dax3)
		self._callbacks.append(dax4)
		# self._callbacks.append(topo)
		fig.show()
		fig.canvas.draw_idle()
		fig.canvas.start_event_loop(0.001)


	def pop_Qwratio_fig(self, jupyter = False, clim = None, alpha_hillshade = 0.3):

		self.flood.enable_Qwout_recording()

		if(jupyter == False):
			plt.ioff()

		fig,ax = plt.subplots()
		dax = scb.RGridDax(self.grid, ax, alpha_hillshade=1)
		arr = self.flood.get_Qwin()
		if(np.prod(arr.shape) == 0):
			arr = self.grid.zeros()
		else:
			arr.reshape(self.grid.rshp)

		def _cback():
			Qwin = self.flood.get_Qwin()
			mask = Qwin == 0
			tarr = self.flood.get_Qwout_recording()/Qwin
			tarr[mask] = 1.
			return tarr

		Qwax = dax.drape_on(arr, cmap = "RdBu_r", clim = (0.75,1.25), delta_zorder = 1, alpha = 1 - alpha_hillshade, callback = _cback)
		# Qwax = dax.drape_on(arr, cmap = "Blues", clim = None, delta_zorder = 1, alpha = 0.9, callback = self.debugyolo)

		plt.colorbar(Qwax.im, label = "$Q_{wout}/Q_{win}")

		self.active_figs.append(fig)
		self._callbacks.append(dax)
		self._callbacks.append(Qwax)
		fig.show()
		fig.canvas.draw_idle()
		fig.canvas.start_event_loop(0.001)






















































#End of file


import numpy as np
import matplotlib.pyplot as plt
import dagger as dag
import scabbard as scb


class Environment(object):
	"""
	docstring for Environment
	"""
	def __init__(self):
		
		super(Environment, self).__init__()

		self.grid = None
		self.data = None
		self.graph = None
		self.connector = None
		self.graphflood = None
		self.param = dag.ParamBag()

	def init_connector(self):
		'''
		TODO:: add aprameters to set up boundary conditions and all
		'''
		self.connector.init()
		self.connector.compute()



	def init_GF2(self):
		'''
		TODO:: WRITE THE TO DO
		'''
		self.graphflood = dag.GF2(self.connector,0,self.data, self.param)
		self.graphflood.init()


	def change_BCs(self, BCs):
		self.data.set_boundaries(BCs.ravel())
		self.connector = dag.Connector8(self.grid.nx, self.grid.ny, self.grid.dx, self.grid.dy, self.data)
		self.connector.set_condou(dag.CONBOU.CUSTOM)
		self.connector.init()

	def d(self, data_name = 'surface'):
		'''
		shortcup function for get_data
		'''
		return self.get_data(data_name);

	def get_data(self, data_name = 'hw'):
		'''
		Returns the 2D array corresponding to the data needed
		'''

		if(data_name.lower().replace(' ','') == 'hw'):
			return self.data.get_hw().reshape(self.grid.rshp)
		elif(data_name.replace(' ','') == 'Qw' or data_name.replace(' ','') == 'Qwin'):
			return self.data.get_Qwin().reshape(self.grid.rshp)
		elif(data_name.replace(' ','') == 'Qwout'):
			return self.data.get_Qwout().reshape(self.grid.rshp)
		elif(data_name.replace(' ','') == 'Qs' or data_name.replace(' ','') == 'Qsin'):
			return self.data.get_Qsin().reshape(self.grid.rshp)
		elif(data_name.replace(' ','') == 'Qsout'):
			return self.data.get_Qsout().reshape(self.grid.rshp)
		elif(data_name.replace(' ','') == 'surface'):
			return self.data.get_surface().reshape(self.grid.rshp)
		elif(data_name.replace(' ','') == 'bed_surface'):
			return self.data.get_surface().reshape(self.grid.rshp) if(self.data.get_hw().shape[0] == 0) else (self.data.get_surface() - self.data.get_hw()).reshape(self.grid.rshp)
		else:
			raise ValueError(dataname + " is not recognised as a data identifier");









def env_from_DEM(fname):
	'''
		Load a raster file into an environment with default connector and all
	'''
	env = Environment()
	env.grid = scb.grid.raster2RGrid(fname, np.float64)
	env.data= dag.Hermes()
	env.data.set_surface(env.grid._Z)
	env.connector = dag.Connector8(env.grid.nx, env.grid.ny, env.grid.dx, env.grid.dy, env.data)
	return env


def env_from_slope(
	nx = 512, 
	ny = 512, 
	dx = 5,
	dy = 5,
	z_base = 0,
	slope = 1e-3, 
	noise_magnitude = 1,
	EW = "periodic",
	S = "out",
	N = "force"

	):
	'''
		Generates a slopping grid adn initialise an env with given boundaries to it. 
		It comes with a connector and a graph (todo) alredy pre-prepared for boundaries
	'''

	# Zeros Topo
	Z = np.zeros(ny * nx)

	# Generating the grid
	grid = scb.RGrid(nx, ny, dx, dy, Z)

	# Getting the matrix
	X,Y,Z = grid.XYZ

	Z = - slope * Y
	Z += z_base -  Z.min()

	grid._Z = Z.ravel()

	if(noise_magnitude > 0):
		grid.add_random_noise(0, noise_magnitude)

	bc = np.ones((ny,nx),dtype = np.uint8)
	bc[:,[0,-1]] = 9 if EW == "periodic" else (0 if EW == "noflow" else (4 if EW == "out" else 3)) #9 is periodic, 0 is no flow, 4 is normal out
	bc[0,:] = 8 if N == "force" else (0 if N == "noflow" else (4 if N == "out" else 3))
	bc[-1,:] = 5 if S == "force" else (4 if S == "out" else 3)
	
	if(EW == "noflow" or EW == "periodic"):
		bc[[0,-1],0] = 0
		bc[[0,-1],-1] = 0

	env = Environment()
	env.grid = grid
	env.data = dag.Hermes()
	env.data.set_surface(grid._Z)
	env.data.set_boundaries(bc.ravel())
	env.connector = dag.Connector8(env.grid.nx, env.grid.ny, env.grid.dx, env.grid.dy, env.data)
	env.connector.set_condou(dag.CONBOU.CUSTOM)

	return env


def env_from_array(
	arr,
	nx = 512, 
	ny = 512, 
	dx = 5,
	dy = 5,
	N = 'out',
	S = 'out',
	E = 'out',
	W = 'out',

	):
	'''
		Generates a slopping grid adn initialise an env with given boundaries to it. 
		It comes with a connector and a graph (todo) alredy pre-prepared for boundaries
	'''

	def _get_bound(bound):
		if(bound == "periodic"):
			return 9
		elif(bound == "noflow"):
			return 0
		elif(bound == "out"):
			return 3
		elif(bound == "forcein"):
			return 8
		elif(bound == "forceout"):
			return 5
		elif(bound == "noout"):
			return 6
		else:
			raise ValueError(f"Boundary type needs to be one of ['periodic','noflow','out','forcein','forceout','noout'], but i got {bound}")


	# Zeros Topo
	Z = arr.ravel()

	# Generating the grid
	grid = scb.RGrid(nx, ny, dx, dy, Z)

	bc = np.ones((ny,nx),dtype = np.uint8)
	bc[:,0] = _get_bound(W)
	bc[:,-1] = _get_bound(E)
	bc[0,:] = _get_bound(N)
	bc[-1,:] = _get_bound(S)
	
	if(E == "noflow" or E == "periodic" or W == "noflow" or W == "periodic"):
		bc[[0,-1],0] = 0
		bc[[0,-1],-1] = 0

	env = Environment()
	env.grid = grid
	env.data = dag.Hermes()
	env.data.set_surface(grid._Z)
	env.data.set_boundaries(bc.ravel())
	env.connector = dag.Connector8(env.grid.nx, env.grid.ny, env.grid.dx, env.grid.dy, env.data)
	env.connector.set_condou(dag.CONBOU.CUSTOM)

	return env








import numba as nb
import numpy as np


@nb.njit()
def mask_watershed_SFD(strat_node, Stack ,Sreceivers):
	'''
	mask all the nodes draining to a single one following an already calculated single flow graph
	'''

	mask = np.zeros_like(Sreceivers, dtype = np.uint8)

	mask[strat_node] = 1

	for node in Stack:
		if(node == strat_node):
			continue
			
		mask[node] = mask[Sreceivers[node]]

	# As simple as that
	return mask





'''
This is a first attempt to make a map automator.
Probably will just test different way here before making a better one
B.G
'''

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from scabbard import Dax

# My cat just walked on the keyboard, let's keep that
# p0----------ppppppppppppppppppppppppp+699999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999999990


class Dfig(object):

	"""docstring for Dfig"""

	def __init__(self, fig = None, axes = None):

		super(Dfig, self).__init__()

		self.fig = None
		self.axes = {}

		if(fig == None):
			self.init_default_fig()
		else:

			self.fig = fig

			if( isinstance(axes,list)):
				for ax in axes:
					self.axes[ax.key] = ax
			else:
				self.axes[axes.key] = axes
				
	def init_default_fig(self):
		"""
		Naive init of default figure
		"""
		fig, ax = plt.subplots()
		self.fig = fig
		tdax = Dax(ax)
		self.axes = {}
		self.axes[tdax.key] =  tdax

	def update(self):
		for ax in self.axes:
			ax.update()
		self.fig.canvas.draw()
			

# class Map(object):

# 	"""
# 		docstring for Map
# 	"""
	
# 	def __init__(self, fig, ax):
# 		super(Map, self).__init__()
# 		self.fig = fig
# 		self.ax
		



















































# end of file


from importlib import resources 
import numpy as np
import numba as nb
import scabbard as scb

import subprocess as sub

# Path to your JSON file
with resources.open_text('scabbard', '_blandplot.py') as config_file:
	blender_script = config_file.name



@nb.njit()
def _fill_arrays(data,displacement_array,color_array, intensity = 0.7):
    k=0
    for j in range(0,data.shape[1]+2):
        for i in range(0,data.shape[0]+2):
            i_data = i - 1
            j_data = j - 1

            i_minus = max(i_data-1,0)
            i_plus = min(i_data+1,data.shape[0]-1)
            j_minus = max(j_data-1,0)
            j_plus = min(j_data+1,data.shape[1]-1)
            
            i_data = min(max(i_data,0),data.shape[0]-1)
            j_data = min(max(j_data,0),data.shape[1]-1)
            
            if i >= 1 and j >= 1 and i <= data.shape[0] and j <= data.shape[1]:
                if np.isnan(data[i_data,j_data]):
                    displacement_array[k:k+4] = np.nan
                    color_array[k:k+4] = 1.0
                else:
                    displacement_array[k:k+4] = data[i_data,j_data]
                    # Get the RGBA color for the specified value
    #                rgba_color = np.array(sm.to_rgba(data[i_data,j_data]))
    #                rgba_color *= luminosity_scale
                    
                    color_array[k:k+4] = data[i_data,j_data] * intensity
                    # color_array[k:k+3] *= luminosity_scale
    #                color_array[k:k+4] = rgba_color
            else:
                displacement_array[k:k+4] = np.nan
                color_array[k:k+4] = 1.0
            k+=4
            
@nb.njit()
def _fill_arrays_cc(data, colors,displacement_array,color_array, intensity = 0.7):
    k=0
    for j in range(0,data.shape[1]+2):
        for i in range(0,data.shape[0]+2):
            i_data = i - 1
            j_data = j - 1

            i_minus = max(i_data-1,0)
            i_plus = min(i_data+1,data.shape[0]-1)
            j_minus = max(j_data-1,0)
            j_plus = min(j_data+1,data.shape[1]-1)
            
            i_data = min(max(i_data,0),data.shape[0]-1)
            j_data = min(max(j_data,0),data.shape[1]-1)
            
            if i >= 1 and j >= 1 and i <= data.shape[0] and j <= data.shape[1]:
                if np.isnan(data[i_data,j_data]):
                    displacement_array[k:k+4] = np.nan
                    color_array[k:k+4] = 1.0
                else:
                    displacement_array[k:k+4] = data[i_data,j_data]
                    # Get the RGBA color for the specified value
    #                rgba_color = np.array(sm.to_rgba(data[i_data,j_data]))
    #                rgba_color *= luminosity_scale
                    
                    color_array[k:k+4] = colors[i_data,j_data] 
                    # color_array[k:k+3] *= luminosity_scale
    #                color_array[k:k+4] = rgba_color
            else:
                displacement_array[k:k+4] = np.nan
                color_array[k:k+4] = 1.0
            k+=4
            



def plot_blender_from_array(_data, 
	fprefix = "__blend",
	save_prefix = "render",
	dx = 5.,
	gpu = True,
	perspective = True,
	ortho_scale = 2.,
	focal_length = 40,
	f_stop = 50000,
	shiftx = 0.,
	shifty = 0.,
	camera_tilt =  45.0,
	camera_rotation =  0.,
	sun_tilt =  25.0,
	sun_rotation =  315.0,
	sun_intensity =  0.2,
	exaggeration =  1.5,
	recast_minZ =  0.,
	recast_maxZ =  1.,
	number_of_subdivisions = 1000,
	res_x = 1920,
	res_y = 1080,
	samples = 50,
	intensity = 0.7,
	custom_colors = None

	):
	
	data = np.copy(_data)

	# create an image that holds the data
	displacement_array = np.zeros(((data.shape[0]+2)*(data.shape[1]+2)*4), dtype=np.float32)
	color_array = np.zeros(((data.shape[0]+2)*(data.shape[1]+2)*4), dtype=np.float32)

	# find minimum elevation
	data_min = np.min(data[data!=-9999])
	# find maximum elevation
	data_max = np.max(data[data!=-9999])
	# calculate total relief
	relief = data_max - data_min
	#scale values from 0 to 1
	data[data!=-9999] = (data[data!=-9999] - data_min) / relief
	# recasting the min max values
	data[data!=-9999] = (data[data!=-9999] - recast_minZ) / (recast_maxZ - recast_minZ)

	_fill_arrays(data,displacement_array,color_array, intensity) if custom_colors is None else _fill_arrays_cc(data, custom_colors, displacement_array, color_array, intensity)



	np.save(fprefix + ".npy", data.astype(np.float32))
	np.save(fprefix + "_displacement_array.npy", displacement_array)
	np.save(fprefix + "_color_array.npy", color_array)


	cmd = ""
	cmd += scb.config.query('blender')
	cmd += "  --background --python " + blender_script + " "
	cmd += "--gpu " if gpu else ""
	cmd += "--perspective " if perspective else ""
	cmd += f"--fprefix '{fprefix}' "
	cmd += f"--save_prefix '{save_prefix}' "
	cmd += f"--ortho_scale {ortho_scale} "
	cmd += f"--focal_length {focal_length} "
	cmd += f"--f_stop {f_stop} "
	cmd += f"--dx {dx} "
	cmd += f"--shiftx {shiftx} "
	cmd += f"--shifty {shifty} "
	cmd += f"--camera_tilt {camera_tilt} "
	cmd += f"--camera_rotation {camera_rotation} "
	cmd += f"--sun_tilt {sun_tilt} "
	cmd += f"--sun_rotation {sun_rotation} "
	cmd += f"--sun_intensity {sun_intensity} "
	cmd += f"--exaggeration {exaggeration} "
	cmd += f"--recast_minZ {recast_minZ} "
	cmd += f"--recast_maxZ {recast_maxZ} "
	cmd += f"--number_of_subdivisions {number_of_subdivisions} "
	cmd += f"--res_x {res_x} "
	cmd += f"--res_y {res_y} "
	cmd += f"--samples {samples} "
	cmd += f"--relief {relief} "



	sub.run(cmd, shell = True, check = False)





	# parser.add_argument('-G', '--gpu', action='store_true', help='GPU acceleration')
	# parser.add_argument('-P', '--perspective', action='store_true', help='Camera set to perspective or orthogonal')
	# parser.add_argument('-f', '--fprefix', type=str, help='file name')
	# parser.add_argument('-o', '--save_prefix', type=str, help='file name')
	# parser.add_argument('--ortho_scale', type=float, default = 2., help='file name')
	# parser.add_argument('--focal_length', type=float, default = 40., help='file name')
	# parser.add_argument('--f_stop', type=float, default = 500000., help='file name')
	# parser.add_argument('--dx', type=float, default = 30., help='file name')
	# parser.add_argument('--shiftx', type=float, default = 0., help='file name')
	# parser.add_argument('--shifty', type=float, default = 0., help='file name')
	# parser.add_argument('--camera_tilt', type = float, default = 45.0, help = '')
	# parser.add_argument('--camera_rotation', type = float, default = 0., help = '')
	# parser.add_argument('--sun_tilt', type = float, default = 25.0, help = '')
	# parser.add_argument('--sun_rotation', type = float, default = 315.0, help = '')
	# parser.add_argument('--sun_intensity', type = float, default = 0.2, help = '')
	# parser.add_argument('--exaggeration', type = float, default = 1.5, help = '')
	# parser.add_argument('--recast_minZ', type = float, default = 0., help = '')
	# parser.add_argument('--recast_maxZ', type = float, default = 1.4, help = '')
	# parser.add_argument('--number_of_subdivisions', type = int, default = 1000 , help = '')
	# parser.add_argument('--res_x', type = int, default = 1920 , help = '')
	# parser.add_argument('--res_y', type = int, default = 1080 , help = '')
	# parser.add_argument('--samples', type = int, default = 50 , help = '')



'''
Manages the geogrphical conversion and all
WILL EVOLVE QUITE A LOT
Not sure I know how I want ot handle that ATM
'''


class geog:
	'''
	Simple class wrapping the geographical information
	Will evolve
	'''

	def __init__(self, 
		xmin = 0., 
		ymin = 0.,
		xmax = 0.,
		ymax = 0.,
		crs = None

	):
		
		self.xmin = xmin
		self.ymin = ymin
		self.xmax = xmax
		self.ymax = ymax
		self.crs = crs





'''
These scripts provide 
'''


import matplotlib.pyplot as plt
import numpy as np
import scabbard as scb



def _legacy_hillshaded_basemap(dem, sea_level = None, fig = None, ax = None, **kwargs):
	'''
	Ongoing deprecation: switching to the new grid system
	Return a fig,ax with a hillshaded relief to the right extent. Ready to be used as base map to plot something on the top

	Arguments:
		- dem: the Rgrid/DEM object with the topography
		- sea_level: if not None, remove everything < sea_level
		- **kwargs: any arguments that would go with matplotlib.pyplot.subplots

	Returns:
		- fig,ax objects

	Authors:
		- B.G.

	'''

	# Creating the figure
	if(fig is None):
		fig,ax = plt.subplots(1, 1, **kwargs)

	# array to plot
	tp = scb.rvd.std_hillshading(dem.Z2D, direction = 40., inclinaison = 55., exaggeration = 1.2, use_gpu = False, D4 = True, dx = dem.dx) + scb.rvd.std_hillshading(dem.Z2D, direction = 85., inclinaison = 55., exaggeration = 1.2, use_gpu = False, D4 = True, dx = dem.dx)
	
	tp /= 2
	
	if not (sea_level is None):
		tp[dem.Z2D<sea_level] = np.nan

	ax.imshow(
		tp,
		cmap = 'gray',
		extent = dem.extent()
		)

	ax.set_xlabel("Easting (m)")
	ax.set_ylabel("Northing (m)")
	
	return fig, ax



def hillshaded_basemap(dem, sea_level = None, fig = None, ax = None, use_gpu = False, mask = None, **kwargs):
	'''
	Return a fig,ax with a hillshaded relief to the right extent. Ready to be used as base map to plot something on the top

	Arguments:
		- dem: raster object with the topography, will be redirected to the right function if former Rgrid format
		- sea_level: if not None, remove everything < sea_level
		- **kwargs: any arguments that would go with matplotlib.pyplot.subplots

	Returns:
		- fig,ax objects

	Authors:
		- B.G.

	'''

	# Legacy compatibility layer
	if(isinstance(dem, scb.RGrid)):
		return _legacy_hillshaded_basemap(dem, sea_level = sea_level,fig = fig, ax = ax, **kwargs)

	# Creating the figure
	if(fig is None):
		fig,ax = plt.subplots(1, 1, **kwargs)

	# array to plot
	tp = scb.rvd.std_hillshading(dem.Z, 
		direction = 40., inclinaison = 55., exaggeration = 1.2, 
		use_gpu = use_gpu, D4 = False, dx = dem.geo.dx) + scb.rvd.std_hillshading(dem.Z, 
		direction = 85., inclinaison = 55., exaggeration = 1.2, use_gpu = use_gpu,
		 D4 = False, dx = dem.geo.dx)
	
	tp /= 2
	
	if not (sea_level is None):
		tp[dem.Z<sea_level] = np.nan

	if not (mask is None):
		tp[mask] = np.nan

	ax.imshow(
		tp,
		cmap = 'gray',
		extent = dem.geo.extent
		)

	ax.set_xlabel("Easting (m)")
	ax.set_ylabel("Northing (m)")


	return fig, ax


def hs_drape(dem, arr2D, cmap = 'cividis', label = 'Metrics', alpha = 0.6, 
	cut_off_min = None, cut_off_max = None, sea_level = None, vmin = None,
	vmax = None, res = None, fig = None, ax = None, mask = None, kwargs_imshow = None, use_gpu = False, **kwargs):
	'''
	Quick visualisation of a DEM as a hillshade + an imshow-like data on the top of it with the same extent

	Arguments:
		- dem: the topographic grid
		- arr2D: the array to be draped, to the extent of the dem
		- cmap: the colormap to use
		- label: label of the colorbar
		- alpha: the transparency of the arr2D
		- cut_off_min/max: any data on arr2D below or above this value will be ignored
		- sea_level: any data on topo and arr2D where topo < that value will be ignored
		- vmin/vmax: the extents of the colors for the cmap of arr2D
		- mask: a binary mask to set additional nodata sections manually
		- **kwargs: anything that goes into fig, ax = plt.subplots(...)
	Returns:
		- fig,ax  with everything plotted on it

	Author:
		- B.G. (last modification: 08/2024)
	'''

	if(fig is None):
		fig, ax = hillshaded_basemap(dem, sea_level = sea_level, mask = mask,use_gpu = use_gpu ,**kwargs)
	else:
		hillshaded_basemap(dem, sea_level = sea_level, fig = fig, ax = ax, mask = mask,use_gpu = use_gpu, **kwargs)

	if(isinstance(dem, scb.RGrid)):
		legacy = True
	else:
		legacy = False

	tp = arr2D.copy()
	if not (cut_off_min is None):
		tp[arr2D<cut_off_min] = np.nan
	if not (cut_off_max is None):
		tp[arr2D>cut_off_max] = np.nan

	if not (sea_level is None ):
		tp[(dem.Z2D <= sea_level) if legacy else (dem.Z <= sea_level) ] = np.nan

	if not (mask is None ):
		tp[mask] = np.nan
	

	if(kwargs_imshow is None):
		im = ax.imshow(tp, extent = dem.extent() if legacy else dem.geo.extent, cmap = cmap, alpha = alpha, vmin = vmin, vmax = vmax)
	else:
		im = ax.imshow(tp, extent = dem.extent() if legacy else dem.geo.extent, **kwargs_imshow)

	if(isinstance(res,dict)):
		res['fig'] = fig
		res['ax'] = ax
		res['im'] = im


	return fig,ax



import taichi as ti
from scabbard.riverdale.rd_grid import GRID
import scabbard.riverdale.rd_grid as gridfuncs
import scabbard.riverdale.rd_helper_surfw as hsw



@ti.kernel
def compute_D4(Z:ti.template(), D4dir:ti.template(), BCs:ti.template()):
	'''
	Experimental tests on drainage area calculations
	Do not use at the moment
	B.G.
	'''

	# Traversing each nodes
	for i,j in Z:

		D4dir[i,j] = ti.uint8(5)

		# If the node cannot give and can only receive, I pass this node
		if(gridfuncs.can_give(i,j,BCs) == False or gridfuncs.is_active(i,j,BCs) == False):
			continue

		# Keeping in mind the steepest slope in the x and y direction to calculate the norm of the vector
		SS = 0.
		checked = True

	
		# Traversing Neighbours
		for k in range(4):
			# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
			ir,jr = gridfuncs.neighbours(i,j,k, BCs)

			# if not a neighbours, by convention is < 0 and I pass
			if(ir == -1):
				continue

			# Local hydraulic slope
			tS = Z[i,j] - Z[ir,jr]
			tS /= GRID.dx

			# If < 0, neighbour is a donor and I am not interested
			if(tS <= 0):
				continue

			# Registering the steepest clope in both directions (see rd_grid header lines for erxplanation on the standard)
			if(tS > SS):
				D4dir[i,j] = ti.uint8(k)
				SS = tS


@ti.kernel
def compute_D4_Zw(Z:ti.template(), hw:ti.template(), D4dir:ti.template(), BCs:ti.template()):
	'''
	Experimental tests on drainage area calculations
	Do not use at the moment
	B.G.
	'''

	# Traversing each nodes
	for i,j in Z:

		D4dir[i,j] = ti.uint8(5)

		# If the node cannot give and can only receive, I pass this node
		if(gridfuncs.can_give(i,j,BCs) == False or gridfuncs.is_active(i,j,BCs) == False):
			continue

		# Keeping in mind the steepest slope in the x and y direction to calculate the norm of the vector
		SS = 0.
		checked = True

	
		# Traversing Neighbours
		for k in range(4):
			# getting neighbour k (see rd_grid header lines for erxplanation on the standard)
			ir,jr = gridfuncs.neighbours(i,j,k, BCs)

			# if not a neighbours, by convention is < 0 and I pass
			if(ir == -1):
				continue

			# Local hydraulic slope
			tS = hsw.Zw(Z,hw,i,j) - hsw.Zw(Z,hw,ir,jr)
			tS /= GRID.dx

			# If < 0, neighbour is a donor and I am not interested
			if(tS <= 0):
				continue

			# Registering the steepest clope in both directions (see rd_grid header lines for erxplanation on the standard)
			if(tS > SS):
				D4dir[i,j] = ti.uint8(k)
				SS = tS


'''
Sets of function to compute hydrodynamics with RiverDale

B.G. - 29/04/2024
'''

import taichi as ti
import numpy as np
from enum import Enum
import scabbard.utils as scaut 
from scabbard.riverdale.rd_grid import GRID
import scabbard.riverdale.rd_grid as gridfuncs
import scabbard.steenbok as ste
import math
import numba as nb


@ti.kernel
def _hillshading(Z:ti.template(), hillshade:ti.template(), BCs:ti.template(), light_dir: ti.f32, light_elev: ti.f32, Z_exaggeration:ti.f32, inverted:bool, dx:ti.f32):
	'''
	🛠🛠🛠 core kernel for producing a nice hillshade

	Arguments:
		- Z: the 2D field of topographic surface
		- hillshade: the 2D field of hillshade to be computed in place
		- BCs: 2D field of boundary conditions
		- light_dir: 0 - 360,  direction of the light
		- light_elev: 0 - 90 inclinaison ("elevation") of the light
		- Z_exaggeration: higher values = higher contrast. Neutral is 1.
		- inverted: True -> "natural" hillshade, False -> "crayon style"

	Returns:
		- Nothing, edits hillshade in place

	Authors:
		- B.G. (last modifications: 06/2024) 
	'''

	# Yolo yolo beng beng
	for i, j in Z:

		# defaulting to no shade
		shaded = 0.

		# Ignoring nodes inactive (e.g. no data and all)
		if gridfuncs.is_active(i,j,BCs):

			# Get elevation differences between D4 neighbors
			dzdx = (Z[i + 1, j] - Z[i - 1, j]) * Z_exaggeration / (2.0 * dx)
			dzdy = (Z[i, j + 1] - Z[i, j - 1]) * Z_exaggeration / (2.0 * dx)
			
			# Calculate slope and aspect
			slope = ti.atan2(ti.sqrt(dzdx**2 + dzdy**2),1)
			aspect = ti.atan2(dzdy, dzdx)
			
			# Calculate illumination angle
			zenith = ti.math.radians(90 - light_elev)
			azimuth = ti.math.radians(light_dir)
			
			# Calculate the hillshade value
			shaded = ti.cos(zenith) * ti.cos(slope) + ti.sin(zenith) * ti.sin(slope) * ti.cos(azimuth - aspect)
			# shaded = ti.max(0, shaded)  # Ensure non-negative
			shaded = ti.abs(shaded)  # Ensure non-negative

		# Saving the value aaaaaand
		hillshade[i, j] = shaded if(inverted) else (1 - shaded)

	# Done


@ti.kernel
def _std_hillshading(Z:ti.template(), hillshade:ti.template(), light_dir: ti.f32, light_elev: ti.f32, Z_exaggeration:ti.f32, inverted:bool, dx:ti.f32):
	'''
	🛠🛠🛠 core kernel for producing a nice hillshade

	Arguments:
		- Z: the 2D field of topographic surface
		- hillshade: the 2D field of hillshade to be computed in place
		- BCs: 2D field of boundary conditions
		- light_dir: 0 - 360,  direction of the light
		- light_elev: 0 - 90 inclinaison ("elevation") of the light
		- Z_exaggeration: higher values = higher contrast. Neutral is 1.
		- inverted: True -> "natural" hillshade, False -> "crayon style"

	Returns:
		- Nothing, edits hillshade in place

	Authors:
		- B.G. (last modifications: 06/2024) 
	'''

	# Yolo yolo beng beng
	for i, j in Z:

		# defaulting to no shade
		shaded = 0.

		# Ignoring nodes inactive (e.g. no data and all)
		if i > 0 and i < Z.shape[0] - 1 and j>0 and j < Z.shape[1] - 1:

			# Get elevation differences between D4 neighbors
			dzdx = (Z[i + 1, j] - Z[i - 1, j]) * Z_exaggeration / (2.0 * dx)
			dzdy = (Z[i, j + 1] - Z[i, j - 1]) * Z_exaggeration / (2.0 * dx)
			
			# Calculate slope and aspect
			slope = ti.atan2(ti.sqrt(dzdx**2 + dzdy**2),1)
			aspect = ti.atan2(dzdy, dzdx)
			
			# Calculate illumination angle
			zenith = ti.math.radians(90 - light_elev)
			azimuth = ti.math.radians(light_dir)
			
			# Calculate the hillshade value
			shaded = ti.cos(zenith) * ti.cos(slope) + ti.sin(zenith) * ti.sin(slope) * ti.cos(azimuth - aspect)
			shaded = ti.max(0, shaded)  # Ensure non-negative

		# Saving the value aaaaaand
		hillshade[i, j] = shaded if(inverted) else (1 - shaded)

@nb.njit()
def _std_hillshading_cpu_D4(Z, light_dir, light_elev, Z_exaggeration, inverted:bool, dx:float):
	'''
	🛠🛠🛠 core kernel for producing a nice hillshade

	Arguments:
		- Z: the 2D field of topographic surface
		- hillshade: the 2D field of hillshade to be computed in place
		- BCs: 2D field of boundary conditions
		- light_dir: 0 - 360,  direction of the light
		- light_elev: 0 - 90 inclinaison ("elevation") of the light
		- Z_exaggeration: higher values = higher contrast. Neutral is 1.
		- inverted: True -> "natural" hillshade, False -> "crayon style"

	Returns:
		- Nothing, edits hillshade in place

	Authors:
		- B.G. (last modifications: 06/2024) 
	'''

	hillshade = np.zeros_like(Z)
	BCs = np.ones(Z.shape).astype(np.uint8)
	ny,nx = BCs.shape
	

	# Yolo yolo beng beng
	for i in range(Z.shape[0]):
		for j in range(Z.shape[1]):

			# defaulting to no shade
			shaded = 0.

			# Ignoring nodes inactive (e.g. no data and all)
			if i > 0 and i < Z.shape[0] - 1 and j>0 and j < Z.shape[1] - 1:

				bottom = ste.neighbours_D4(i, j, 3, BCs, nx, ny)
				top = ste.neighbours_D4(i, j, 0, BCs, nx, ny)
				left = ste.neighbours_D4(i, j, 1, BCs, nx, ny)
				right = ste.neighbours_D4(i, j, 2, BCs, nx, ny)

				# Get elevation differences between D4 neighbors
				dzdx = (Z[bottom[0],bottom[1]] - Z[top[0],top[1]]) * Z_exaggeration / (2*dx)
				dzdy = (Z[right[0],right[1]] - Z[left[0],left[1]]) * Z_exaggeration / (2*dx)
				
				# Calculate slope and aspect
				slope = math.atan2(math.sqrt(dzdx**2 + dzdy**2),1)
				aspect = math.atan2(dzdy, dzdx)
				
				# Calculate illuminamathon angle
				zenith = math.radians(90 - light_elev)
				azimuth = math.radians(light_dir)
				
				# Calculate the hillshade value
				shaded = math.cos(zenith) * math.cos(slope) + math.sin(zenith) * math.sin(slope) * math.cos(azimuth - aspect)
				shaded = max(0, shaded)  # Ensure non-negative

			# Saving the value aaaaaand
			hillshade[i, j] = shaded if(inverted) else (1 - shaded)

	return hillshade


@nb.njit()
def _std_hillshading_cpu_D8(Z, light_dir, light_elev, Z_exaggeration, inverted:bool, dx:float):
	'''
	🛠🛠🛠 core kernel for producing a nice hillshade

	Arguments:
		- Z: the 2D field of topographic surface
		- hillshade: the 2D field of hillshade to be computed in place
		- BCs: 2D field of boundary conditions
		- light_dir: 0 - 360,  direction of the light
		- light_elev: 0 - 90 inclinaison ("elevation") of the light
		- Z_exaggeration: higher values = higher contrast. Neutral is 1.
		- inverted: True -> "natural" hillshade, False -> "crayon style"

	Returns:
		- Nothing, edits hillshade in place

	Authors:
		- B.G. (last modifications: 06/2024) 
	'''

	hillshade = np.zeros_like(Z)
	BCs = np.ones(Z.shape).astype(np.uint8)
	BCs[0,  :] = 3
	BCs[-1, :] = 3
	BCs[:,  0] = 3
	BCs[:, -1] = 3
	ny,nx = BCs.shape
	
	hillshade = hillshade.ravel()
	BCs = BCs.ravel()

	dxy = dx*np.sqrt(2.)

	# Yolo yolo beng beng
	for i in range(Z.ravel().shape[0]):

		# defaulting to no shade
		shaded = 0.

		# Ignoring nodes inactive (e.g. no data and all)
	

		bottomleft = ste.neighbours_D8_flat(i, 5, BCs, nx, ny)
		bottom = ste.neighbours_D8_flat(i, 6, BCs, nx, ny)
		bottomright = ste.neighbours_D8_flat(i, 7, BCs, nx, ny)
		topleft = ste.neighbours_D8_flat(i, 0, BCs, nx, ny)
		top = ste.neighbours_D8_flat(i, 1, BCs, nx, ny)
		topright = ste.neighbours_D8_flat(i, 2, BCs, nx, ny)
		left = ste.neighbours_D8_flat(i, 3, BCs, nx, ny)
		right = ste.neighbours_D8_flat(i, 4, BCs, nx, ny)

		if bottomleft == -1 or bottom == -1 or bottomright == -1 or topleft == -1 or top == -1 or topright == -1 or left == -1 or right == -1:
			continue

		# Get elevation differences between D4 neighbors
		dzdx = (Z.ravel()[bottom] - Z.ravel()[top]) * Z_exaggeration / (2*dx)
		dzdy = (Z.ravel()[right] - Z.ravel()[left]) * Z_exaggeration / (2*dx)
		dzdxy1 = (Z.ravel()[bottomright] - Z.ravel()[topleft]) * Z_exaggeration / (2*dxy)
		dzdxy2 = (Z.ravel()[bottomleft] - Z.ravel()[topright]) * Z_exaggeration / (2*dxy)
		# print(dzdx)
		
		# Calculate slope and aspect
		slope = np.atan2(np.sqrt(dzdx**2 + dzdy**2),1)
		aspect = np.atan2(dzdy, dzdx)

		# Calculate slope and aspect (diag)
		slope_diag = np.atan2(np.sqrt(dzdxy1**2 + dzdxy2**2),1)
		aspect_diag = np.atan2(dzdy, dzdx)
		
		# Calculate illuminanpon angle
		zenith = np.radians(90 - light_elev)
		azimuth = np.radians(light_dir)

		
		# Calculate the hillshade value
		shaded = np.cos(zenith) * np.cos(slope) + np.sin(zenith) * np.sin(slope) * np.cos(azimuth - aspect)
		shaded_diag = np.cos(zenith) * np.cos(slope_diag) + np.sin(zenith) * np.sin(slope_diag) * np.cos(azimuth - aspect_diag)
		shaded = max(0, (shaded + shaded_diag)/2.)  # Ensure non-negative

		# Saving the value aaaaaand
		hillshade[i] = shaded if(inverted) else (1 - shaded)

	return hillshade.reshape(Z.shape)

def _compute_hillshade(Z, light_dir, light_elev, Z_exaggeration, inverted, dx):
	"""
	Optimized hillshading function using numpy.
	"""
	# Precompute light direction components
	zenith_rad = np.radians(90 - light_elev)
	azimuth_rad = np.radians(light_dir)

	cos_zenith = np.cos(zenith_rad)
	sin_zenith = np.sin(zenith_rad)

	# Compute gradients
	dzdx, dzdy = np.gradient(Z * Z_exaggeration, dx)

	# Compute slope and aspect
	slope = np.arctan(np.sqrt(dzdx**2 + dzdy**2))
	aspect = np.arctan2(dzdy, dzdx)

	# Compute hillshade
	shaded = cos_zenith * np.cos(slope) + sin_zenith * np.sin(slope) * np.cos(azimuth_rad - aspect)
	shaded = np.clip(shaded, 0, 1)  # Ensure non-negative

	return 1 - shaded if inverted else shaded

def hillshading(rd, direction = 315., inclinaison = 45., exaggeration = 4.):
	'''
		🔭🔭🔭 Analysis function
		Hillshades the topography (field Z) of a riverdale's instance.

		Arguments:
			- rd: the riverdale's object
			- direction: 0 - 360, direction of the lighting
			- inclinaison: 0 - 90, inclinaison of the lighting (morning/evening light to noon)
			- exaggeration: higher values means higher contrast

		Returns:
			- A 2D numpy array of hillshade
		Authors:
			- B.G. (last modifications: 06/2024)
	'''

	# Prefetching the field
	hillshade, = rd.query_temporary_fields(1,dtype = ti.f32)

	# Running the GPU kernel
	_hillshading(rd.Z,hillshade,rd.BCs, direction, inclinaison, exaggeration, True)

	# Done, copying back to cpu and returning
	return hillshade.to_numpy()



def std_hillshading(Z2D, direction = 315., inclinaison = 45., exaggeration = 4., use_gpu = False, D4 = False, dx = 1.):
	'''
		🔭🔭🔭 Analysis function
		Hillshades the topography (field Z) of a riverdale's instance.

		Arguments:
			- rd: the riverdale's object
			- direction: 0 - 360, direction of the lighting
			- inclinaison: 0 - 90, inclinaison of the lighting (morning/evening light to noon)
			- exaggeration: higher values means higher contrast

		Returns:
			- A 2D numpy array of hillshade
		Authors:
			- B.G. (last modifications: 06/2024)
	'''
	
	if(use_gpu):
		try:
			ti.init(ti.gpu)
		except:
			ti.init(ti.cpu)

		# Prefetching the field
		hillshade = ti.field(dtype = ti.f32, shape = Z2D.shape)
		tZ = ti.field(dtype = ti.f32, shape = Z2D.shape)
		tZ.from_numpy(Z2D)

		# Running the GPU kernel
		_std_hillshading(tZ,hillshade, direction, inclinaison, exaggeration, True, dx)

		# Done, copying back to cpu and returning
		ret =  hillshade.to_numpy()
		ti.init()
		return ret

	else:

		return _compute_hillshade(Z2D, direction, inclinaison, exaggeration, False, dx) if D4 else _compute_hillshade(Z2D, direction, inclinaison, exaggeration, False, dx)




dark_orange = '''
QToolTip
{
     border: 1px solid black;
     background-color: #ffa02f;
     padding: 1px;
     border-radius: 3px;
     opacity: 100;
}

QWidget
{
    color: #b1b1b1;
    background-color: #323232;
}

QWidget:item:hover
{
    background-color: QLinearGradient( x1: 0, y1: 0, x2: 0, y2: 1, stop: 0 #ffa02f, stop: 1 #ca0619);
    color: #000000;
}

QWidget:item:selected
{
    background-color: QLinearGradient( x1: 0, y1: 0, x2: 0, y2: 1, stop: 0 #ffa02f, stop: 1 #d7801a);
}

QMenuBar::item
{
    background: transparent;
}

QMenuBar::item:selected
{
    background: transparent;
    border: 1px solid #ffaa00;
}

QMenuBar::item:pressed
{
    background: #444;
    border: 1px solid #000;
    background-color: QLinearGradient(
        x1:0, y1:0,
        x2:0, y2:1,
        stop:1 #212121,
        stop:0.4 #343434/*,
        stop:0.2 #343434,
        stop:0.1 #ffaa00*/
    );
    margin-bottom:-1px;
    padding-bottom:1px;
}

QMenu
{
    border: 1px solid #000;
}

QMenu::item
{
    padding: 2px 20px 2px 20px;
}

QMenu::item:selected
{
    color: #000000;
}

QWidget:disabled
{
    color: #404040;
    background-color: #323232;
}

QAbstractItemView
{
    background-color: QLinearGradient( x1: 0, y1: 0, x2: 0, y2: 1, stop: 0 #4d4d4d, stop: 0.1 #646464, stop: 1 #5d5d5d);
}

QWidget:focus
{
    /*border: 2px solid QLinearGradient( x1: 0, y1: 0, x2: 0, y2: 1, stop: 0 #ffa02f, stop: 1 #d7801a);*/
}

QLineEdit
{
    background-color: QLinearGradient( x1: 0, y1: 0, x2: 0, y2: 1, stop: 0 #4d4d4d, stop: 0 #646464, stop: 1 #5d5d5d);
    padding: 1px;
    border-style: solid;
    border: 1px solid #1e1e1e;
    border-radius: 5;
}

QPushButton
{
    color: #b1b1b1;
    background-color: QLinearGradient( x1: 0, y1: 0, x2: 0, y2: 1, stop: 0 #565656, stop: 0.1 #525252, stop: 0.5 #4e4e4e, stop: 0.9 #4a4a4a, stop: 1 #464646);
    border-width: 1px;
    border-color: #1e1e1e;
    border-style: solid;
    border-radius: 6;
    padding: 3px;
    font-size: 12px;
    padding-left: 5px;
    padding-right: 5px;
}

QPushButton:pressed
{
    background-color: QLinearGradient( x1: 0, y1: 0, x2: 0, y2: 1, stop: 0 #2d2d2d, stop: 0.1 #2b2b2b, stop: 0.5 #292929, stop: 0.9 #282828, stop: 1 #252525);
}

QComboBox
{
    selection-background-color: #ffaa00;
    background-color: QLinearGradient( x1: 0, y1: 0, x2: 0, y2: 1, stop: 0 #565656, stop: 0.1 #525252, stop: 0.5 #4e4e4e, stop: 0.9 #4a4a4a, stop: 1 #464646);
    border-style: solid;
    border: 1px solid #1e1e1e;
    border-radius: 5;
}

QComboBox:hover,QPushButton:hover
{
    border: 2px solid QLinearGradient( x1: 0, y1: 0, x2: 0, y2: 1, stop: 0 #ffa02f, stop: 1 #d7801a);
}


QComboBox:on
{
    padding-top: 3px;
    padding-left: 4px;
    background-color: QLinearGradient( x1: 0, y1: 0, x2: 0, y2: 1, stop: 0 #2d2d2d, stop: 0.1 #2b2b2b, stop: 0.5 #292929, stop: 0.9 #282828, stop: 1 #252525);
    selection-background-color: #ffaa00;
}

QComboBox QAbstractItemView
{
    border: 2px solid darkgray;
    selection-background-color: QLinearGradient( x1: 0, y1: 0, x2: 0, y2: 1, stop: 0 #ffa02f, stop: 1 #d7801a);
}

QComboBox::drop-down
{
     subcontrol-origin: padding;
     subcontrol-position: top right;
     width: 15px;

     border-left-width: 0px;
     border-left-color: darkgray;
     border-left-style: solid; /* just a single line */
     border-top-right-radius: 3px; /* same radius as the QComboBox */
     border-bottom-right-radius: 3px;
 }

QComboBox::down-arrow
{
     image: url(:/down_arrow.png);
}

QGroupBox:focus
{
border: 2px solid QLinearGradient( x1: 0, y1: 0, x2: 0, y2: 1, stop: 0 #ffa02f, stop: 1 #d7801a);
}

QTextEdit:focus
{
    border: 2px solid QLinearGradient( x1: 0, y1: 0, x2: 0, y2: 1, stop: 0 #ffa02f, stop: 1 #d7801a);
}

QScrollBar:horizontal {
     border: 1px solid #222222;
     background: QLinearGradient( x1: 0, y1: 0, x2: 0, y2: 1, stop: 0.0 #121212, stop: 0.2 #282828, stop: 1 #484848);
     height: 7px;
     margin: 0px 16px 0 16px;
}

QScrollBar::handle:horizontal
{
      background: QLinearGradient( x1: 0, y1: 0, x2: 1, y2: 0, stop: 0 #ffa02f, stop: 0.5 #d7801a, stop: 1 #ffa02f);
      min-height: 20px;
      border-radius: 2px;
}

QScrollBar::add-line:horizontal {
      border: 1px solid #1b1b19;
      border-radius: 2px;
      background: QLinearGradient( x1: 0, y1: 0, x2: 1, y2: 0, stop: 0 #ffa02f, stop: 1 #d7801a);
      width: 14px;
      subcontrol-position: right;
      subcontrol-origin: margin;
}

QScrollBar::sub-line:horizontal {
      border: 1px solid #1b1b19;
      border-radius: 2px;
      background: QLinearGradient( x1: 0, y1: 0, x2: 1, y2: 0, stop: 0 #ffa02f, stop: 1 #d7801a);
      width: 14px;
     subcontrol-position: left;
     subcontrol-origin: margin;
}

QScrollBar::right-arrow:horizontal, QScrollBar::left-arrow:horizontal
{
      border: 1px solid black;
      width: 1px;
      height: 1px;
      background: white;
}

QScrollBar::add-page:horizontal, QScrollBar::sub-page:horizontal
{
      background: none;
}

QScrollBar:vertical
{
      background: QLinearGradient( x1: 0, y1: 0, x2: 1, y2: 0, stop: 0.0 #121212, stop: 0.2 #282828, stop: 1 #484848);
      width: 7px;
      margin: 16px 0 16px 0;
      border: 1px solid #222222;
}

QScrollBar::handle:vertical
{
      background: QLinearGradient( x1: 0, y1: 0, x2: 0, y2: 1, stop: 0 #ffa02f, stop: 0.5 #d7801a, stop: 1 #ffa02f);
      min-height: 20px;
      border-radius: 2px;
}

QScrollBar::add-line:vertical
{
      border: 1px solid #1b1b19;
      border-radius: 2px;
      background: QLinearGradient( x1: 0, y1: 0, x2: 0, y2: 1, stop: 0 #ffa02f, stop: 1 #d7801a);
      height: 14px;
      subcontrol-position: bottom;
      subcontrol-origin: margin;
}

QScrollBar::sub-line:vertical
{
      border: 1px solid #1b1b19;
      border-radius: 2px;
      background: QLinearGradient( x1: 0, y1: 0, x2: 0, y2: 1, stop: 0 #d7801a, stop: 1 #ffa02f);
      height: 14px;
      subcontrol-position: top;
      subcontrol-origin: margin;
}

QScrollBar::up-arrow:vertical, QScrollBar::down-arrow:vertical
{
      border: 1px solid black;
      width: 1px;
      height: 1px;
      background: white;
}


QScrollBar::add-page:vertical, QScrollBar::sub-page:vertical
{
      background: none;
}

QTextEdit
{
    background-color: #242424;
}

QPlainTextEdit
{
    background-color: #242424;
}

QHeaderView::section
{
    background-color: QLinearGradient(x1:0, y1:0, x2:0, y2:1, stop:0 #616161, stop: 0.5 #505050, stop: 0.6 #434343, stop:1 #656565);
    color: white;
    padding-left: 4px;
    border: 1px solid #6c6c6c;
}

QCheckBox:disabled
{
color: #414141;
}

QDockWidget::title
{
    text-align: center;
    spacing: 3px; /* spacing between items in the tool bar */
    background-color: QLinearGradient(x1:0, y1:0, x2:0, y2:1, stop:0 #323232, stop: 0.5 #242424, stop:1 #323232);
}

QDockWidget::close-button, QDockWidget::float-button
{
    text-align: center;
    spacing: 1px; /* spacing between items in the tool bar */
    background-color: QLinearGradient(x1:0, y1:0, x2:0, y2:1, stop:0 #323232, stop: 0.5 #242424, stop:1 #323232);
}

QDockWidget::close-button:hover, QDockWidget::float-button:hover
{
    background: #242424;
}

QDockWidget::close-button:pressed, QDockWidget::float-button:pressed
{
    padding: 1px -1px -1px 1px;
}

QMainWindow::separator
{
    background-color: QLinearGradient(x1:0, y1:0, x2:0, y2:1, stop:0 #161616, stop: 0.5 #151515, stop: 0.6 #212121, stop:1 #343434);
    color: white;
    padding-left: 4px;
    border: 1px solid #4c4c4c;
    spacing: 3px; /* spacing between items in the tool bar */
}

QMainWindow::separator:hover
{

    background-color: QLinearGradient(x1:0, y1:0, x2:0, y2:1, stop:0 #d7801a, stop:0.5 #b56c17 stop:1 #ffa02f);
    color: white;
    padding-left: 4px;
    border: 1px solid #6c6c6c;
    spacing: 3px; /* spacing between items in the tool bar */
}

QToolBar::handle
{
     spacing: 3px; /* spacing between items in the tool bar */
     background: url(:/images/handle.png);
}

QMenu::separator
{
    height: 2px;
    background-color: QLinearGradient(x1:0, y1:0, x2:0, y2:1, stop:0 #161616, stop: 0.5 #151515, stop: 0.6 #212121, stop:1 #343434);
    color: white;
    padding-left: 4px;
    margin-left: 10px;
    margin-right: 5px;
}

QProgressBar
{
    border: 2px solid grey;
    border-radius: 5px;
    text-align: center;
}

QProgressBar::chunk
{
    background-color: #d7801a;
    width: 2.15px;
    margin: 0.5px;
}

QTabBar::tab {
    color: #b1b1b1;
    border: 1px solid #444;
    border-bottom-style: none;
    background-color: #323232;
    padding-left: 10px;
    padding-right: 10px;
    padding-top: 3px;
    padding-bottom: 2px;
    margin-right: -1px;
}

QTabWidget::pane {
    border: 1px solid #444;
    top: 1px;
}

QTabBar::tab:last
{
    margin-right: 0; /* the last selected tab has nothing to overlap with on the right */
    border-top-right-radius: 3px;
}

QTabBar::tab:first:!selected
{
 margin-left: 0px; /* the last selected tab has nothing to overlap with on the right */


    border-top-left-radius: 3px;
}

QTabBar::tab:!selected
{
    color: #b1b1b1;
    border-bottom-style: solid;
    margin-top: 3px;
    background-color: QLinearGradient(x1:0, y1:0, x2:0, y2:1, stop:1 #212121, stop:.4 #343434);
}

QTabBar::tab:selected
{
    border-top-left-radius: 3px;
    border-top-right-radius: 3px;
    margin-bottom: 0px;
}

QTabBar::tab:!selected:hover
{
    /*border-top: 2px solid #ffaa00;
    padding-bottom: 3px;*/
    border-top-left-radius: 3px;
    border-top-right-radius: 3px;
    background-color: QLinearGradient(x1:0, y1:0, x2:0, y2:1, stop:1 #212121, stop:0.4 #343434, stop:0.2 #343434, stop:0.1 #ffaa00);
}

QRadioButton::indicator:checked, QRadioButton::indicator:unchecked{
    color: #b1b1b1;
    background-color: #323232;
    border: 1px solid #b1b1b1;
    border-radius: 6px;
}

QRadioButton::indicator:checked
{
    background-color: qradialgradient(
        cx: 0.5, cy: 0.5,
        fx: 0.5, fy: 0.5,
        radius: 1.0,
        stop: 0.25 #ffaa00,
        stop: 0.3 #323232
    );
}

QCheckBox::indicator{
    color: #b1b1b1;
    background-color: #323232;
    border: 1px solid #b1b1b1;
    width: 9px;
    height: 9px;
}

QRadioButton::indicator
{
    border-radius: 6px;
}

QRadioButton::indicator:hover, QCheckBox::indicator:hover
{
    border: 1px solid #ffaa00;
}

QCheckBox::indicator:checked
{
    image:url(:/images/checkbox.png);
}

QCheckBox::indicator:disabled, QRadioButton::indicator:disabled
{
    border: 1px solid #444;
}
'''


################################################################################
# Import Libraries
import bpy
import numpy as np
import math
from mathutils import Matrix
import importlib
import sys
import os
import argparse


# Create the parser
parser = argparse.ArgumentParser(description='Process some arguments.')

# Add a boolean argument with both short (-b) and long (--flag) flags
# parser.add_argument('-b', "--blenderoptionignore1", action='store_true', help='GPU acceleration')
# parser.add_argument('-P', "--blenderoptionignore2", action='store_true', help='GPU acceleration')
parser.add_argument('--gpu', action='store_true', help='GPU acceleration')
parser.add_argument('--perspective', action='store_true', help='Camera set to perspective or orthogonal')
parser.add_argument('--fprefix', type=str, help='file name')
parser.add_argument('--save_prefix', type=str, help='file name')
parser.add_argument('--ortho_scale', type=float, default = 2., help='file name')
parser.add_argument('--focal_length', type=float, default = 40., help='file name')
parser.add_argument('--f_stop', type=float, default = 500000., help='file name')
parser.add_argument('--dx', type=float, default = 30., help='file name')
parser.add_argument('--shiftx', type=float, default = 0., help='file name')
parser.add_argument('--shifty', type=float, default = 0., help='file name')
parser.add_argument('--camera_tilt', type = float, default = 45.0, help = '')
parser.add_argument('--camera_rotation', type = float, default = 0., help = '')
parser.add_argument('--sun_tilt', type = float, default = 25.0, help = '')
parser.add_argument('--sun_rotation', type = float, default = 315.0, help = '')
parser.add_argument('--sun_intensity', type = float, default = 0.2, help = '')
parser.add_argument('--exaggeration', type = float, default = 1.5, help = '')
parser.add_argument('--recast_minZ', type = float, default = 0., help = '')
parser.add_argument('--recast_maxZ', type = float, default = 1.4, help = '')
parser.add_argument('--number_of_subdivisions', type = int, default = 1000 , help = '')
parser.add_argument('--res_x', type = int, default = 1920 , help = '')
parser.add_argument('--res_y', type = int, default = 1080 , help = '')
parser.add_argument('--samples', type = int, default = 50 , help = '')
parser.add_argument('--relief', type = float, default = 50. , help = '')







args, unknown = parser.parse_known_args()

print(args)
print('\n\n\n')
print(unknown)
print('\n\n\n')


################################################################################
# This sets the render enginer to CYCLES.
# In order for TopoBlender to render your topography correctly, you must use
# this render engine. This engine is optimized for GPUs, so if your computer
# lacks a GPU, TopoBlender may be slow.
GPU_boolean = args.gpu
delete_all = True

# define current scene
scn = bpy.context.scene
if(delete_all):
    # Select all objects in the scene
    bpy.ops.object.select_all(action='SELECT')

# Delete selected objects
bpy.ops.object.delete()

# check the render engine, and change to CYCLES
if not scn.render.engine == 'CYCLES':
    scn.render.engine = 'CYCLES'

#if cycles, change to gpu rendering if user selects GPU
if scn.render.engine == 'CYCLES':
    if GPU_boolean == 1:
        scn.cycles.device = 'GPU'


# Loading the file
data = np.load(args.fprefix + ".npy")
displacement_array = np.load(args.fprefix + "_displacement_array.npy")
color_array = np.load(args.fprefix + "_color_array.npy")

output = args.save_prefix + ".png"

ny,nx = data.shape
dx,dy = args.dx, args.dx

# Standard nodata value
nodata_value = -9999


##########camera parameters##########
#camera type
camera_type = 'perspective' if args.perspective else 'orthogonal' 

#orthogonal
ortho_scale = args.ortho_scale #when using orthogonal scale, increase to "zoom" out


focal_length = args.focal_length #mm when using perspective camera, increase to zoom in
f_stop = args.f_stop # affects depths of field, lower for a shallow dof, higher for wide dof
shift_x = args.shiftx # you may need to shift the camera to center the topo in the frame
shift_y = args.shifty # you may need to shift the camera to center the topo in the frame

#camera location
camera_tilt = args.camera_tilt # #degrees from horizontal
camera_rotation = args.camera_rotation # #camera location degrees CW from North
######################################


##########sun properties##############
sun_tilt = args.sun_tilt # #degrees from horizontal
sun_rotation = args.sun_rotation # #degrees CW from North
sun_intensity = args.sun_intensity # #sun intensity
######################################


#####landscape representation#########
number_of_subdivisions = args.number_of_subdivisions # of subvisions, more increases the detail
exaggeration = args.exaggeration # #vertical exaggeration
recast_minZ = args.recast_minZ #
recast_maxZ = args.recast_maxZ #
luminosity_scale = 0.7
######################################

#########render settings##############
res_x = args.res_x #x resolution of the render
res_y = args.res_y #y resolution of the render
samples = args.samples #number of samples that decides how "good" the render looks. more is better but takes longer
######################################

#convert nodata value to nan
data[data==nodata_value] = np.nan
# data y_length we use this to scale the topography
y_length = (2.0+float(data.shape[1])) * dx


# create an image that holds the data
# displacement_array = np.zeros(((data.shape[0]+2)*(data.shape[1]+2)*4), dtype=np.float32)
# color_array = np.zeros(((data.shape[0]+2)*(data.shape[1]+2)*4), dtype=np.float32)

# # Add a 1-pixel border around the entire topography and rescale the topgraphy
# # so minimum elevation is zero and the maximum elevation is 1.
# _fill_arrays(data,displacement_array,color_array, luminosity_scale)

# Create an image from the datafile        
displacement_image = bpy.data.images.new('displacement_data', data.shape[0]+2, data.shape[1]+2, alpha=False, float_buffer = True)
color_image = bpy.data.images.new('color_data', data.shape[0]+2, data.shape[1]+2, alpha=False, float_buffer = True)
# Fast way to set pixels (since 2.83)
displacement_image.pixels.foreach_set(displacement_array)
color_image.pixels.foreach_set(color_array)
# Pack the image into .blend so it gets saved with it
displacement_image.pack()
color_image.pack()
print("C") 
#make plane
plane_size = 1.0
topo_mesh = bpy.ops.mesh.primitive_plane_add(size=plane_size)
topo_obj = bpy.context.active_object

#Change the shape of the object to match data_aspect
topo_obj.scale = ((2.0+data.shape[0])/(2.0+data.shape[1]),1,1)

#add material
topo_mat = bpy.data.materials.new("topo_mat")
topo_mat.cycles.displacement_method = "DISPLACEMENT"
topo_mat.use_nodes = True

#calculate subdivisions
order_of_magnitude = math.floor(math.log10(number_of_subdivisions))
first_digit = int(np.round(number_of_subdivisions / (10.0 ** order_of_magnitude)))

topo_obj.data.materials.append(topo_mat)
bpy.ops.object.mode_set(mode="EDIT")
for i in range(0,order_of_magnitude):
    bpy.ops.mesh.subdivide(number_cuts=10)
bpy.ops.mesh.subdivide(number_cuts=first_digit)
bpy.ops.object.mode_set(mode="OBJECT")

#add image node - determines the displacement
displacement_image_node = topo_mat.node_tree.nodes.new("ShaderNodeTexImage")
#assign png to image to node
displacement_image_node.image = displacement_image
#change colorspace to b&w
displacement_image_node.image.colorspace_settings.name="Linear FilmLight E-Gamut"

#add image node - determines the color of the lanscape
color_image_node = topo_mat.node_tree.nodes.new("ShaderNodeTexImage")
#assign png to image to node
color_image_node.image = color_image
#change colorspace to b&w
color_image_node.image.colorspace_settings.name="Linear FilmLight E-Gamut"
    
#add displacement node - done
displacement_node = topo_mat.node_tree.nodes.new("ShaderNodeDisplacement")
displacement_node.inputs.get("Scale").default_value = exaggeration * plane_size * args.relief / y_length
displacement_node.inputs.get("Midlevel").default_value = 0.0

#add world sky - done
topo_world = bpy.data.worlds.new('topo_world')
topo_world.use_nodes = True
topo_world_node = topo_world.node_tree.nodes.new("ShaderNodeTexSky")
topo_world_node.sun_elevation = np.radians(sun_tilt)
topo_world_node.sun_rotation = np.radians(sun_rotation-90.0)
topo_world_node.sun_intensity = sun_intensity
topo_world.node_tree.links.new(topo_world_node.outputs['Color'], topo_world.node_tree.nodes['Background'].inputs[0])

#add camera - done
camera_distance = 2.0 #meters
cam = bpy.data.cameras.new('topo_cam')
cam_obj = bpy.data.objects.new('topo_cam',cam)
cam_obj.rotation_euler = (np.radians(90.0 - camera_tilt), np.radians(0), np.radians(270.0 - camera_rotation))
cam_obj.matrix_basis @= Matrix.Translation((0.0, 0.0, camera_distance))

if camera_type == 'orthogonal':
    bpy.data.cameras['topo_cam'].type = 'ORTHO'
    bpy.data.cameras['topo_cam'].ortho_scale = ortho_scale
elif camera_type == 'perspective':
    bpy.data.cameras['topo_cam'].type = 'PERSP'
    bpy.data.cameras['topo_cam'].lens = focal_length
    bpy.data.cameras['topo_cam'].shift_x = shift_x
    bpy.data.cameras['topo_cam'].shift_y = shift_y
    bpy.data.cameras['topo_cam'].dof.use_dof = True
    bpy.data.cameras['topo_cam'].dof.aperture_fstop = f_stop
    bpy.data.cameras['topo_cam'].dof.focus_distance = camera_distance
    
#connect nodes - done
topo_mat.node_tree.links.new(displacement_image_node.outputs["Color"], \
                             displacement_node.inputs["Height"])
topo_mat.node_tree.links.new(displacement_node.outputs["Displacement"], \
                             topo_mat.node_tree.nodes["Material Output"].inputs["Displacement"])
topo_mat.node_tree.links.new(color_image_node.outputs["Color"], \
                             topo_mat.node_tree.nodes["Principled BSDF"].inputs[0])

bpy.context.scene.collection.objects.link(cam_obj)                         
bpy.context.scene.camera = cam_obj    
bpy.context.scene.world = topo_world

#render settings
bpy.context.scene.render.engine = 'CYCLES'
bpy.context.scene.cycles.samples = samples
bpy.context.scene.render.resolution_x = int(res_x)
bpy.context.scene.render.resolution_y = int(res_y)
bpy.context.scene.render.filepath = output
bpy.ops.render.render(write_still=True)


import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import cmcrameri as cm
import string
import random


class Dax(object):
	
	"""
		docstring for Dax
	"""

	def __init__(self, ax, key = None, zorder = 1, fig = None):
		
		super(Dax, self).__init__()

		self.fig = fig
		
		self.ax = ax

		self.zorder = zorder
		
		# // Generating random key if none is given	
		if(key is not None):
			self.key = key
		else:
			self.key = ''.join(random.choices(string.ascii_letters + string.digits, k=8))


	def __hash__(self):
		return hash(self.key)

	def update(self):
		pass


class callbax_image(object):
	
	def __init__(self, im, callback, rshp, clim = None, callback_params = None):
		self.im = im
		self.callback = callback
		self.callback_params = callback_params
		self.rshp = rshp
		self.clim = clim

	def update(self):
		if self.callback_params is None :
			arr = self.callback().reshape(self.rshp)
		else:
			arr = self.callback(*self.callback_params).reshape(self.rshp)
		self.im.set_data(arr)
		if(self.clim is None):
			self.im.set_clim(np.nanmin(arr), np.nanmax(arr))

class callbax_sline(object):
	
	def __init__(self, ax, sline, callback, axylim = None, axylim_ignore = False):
		self.ax = ax
		self.sline = sline[0]
		self.callback = callback
		self.axylim = axylim
		self.axylim_ignore = axylim_ignore

	def update(self):
		
		arr = self.callback()
		self.sline.set_xdata(arr[0])
		self.sline.set_ydata(arr[1])

		if(self.axylim is None and self.axylim_ignore == False):
			perc = 0.05 * (np.nanmax(arr[1]) - np.nanmin(arr[1]))
			self.ax.set_ylim(np.nanmin(arr[1]) - perc, np.nanmax(arr[1]) + perc)
			perc = 0.05 * (np.nanmax(arr[0]) - np.nanmin(arr[0]))
			self.ax.set_xlim(np.nanmin(arr[0]) - perc, np.nanmax(arr[0]) + perc)




class RGridDax(Dax):
	"""docstring for RGridDax"""
	def __init__(self,Rgrid, ax, key = None, cmap = "gist_earth", hillshade = True, alpha_hillshade = 0.45, clim = None, callback_topo = None):
		
		super(RGridDax, self).__init__(ax, key)
		self.grid = Rgrid
		self.base_ax = self.ax.imshow(self.grid.Z2D, extent = self.grid.extent(), cmap = cmap, zorder =  self.zorder)
		self.hillshade_on = hillshade

		self.clim = clim if clim is not None else [self.grid.min(), self.grid.max()]

		if(hillshade):
			self.hillshade_ax = self.ax.imshow(self.grid.hillshade, extent = self.grid.extent(), cmap = "gray", alpha = alpha_hillshade, zorder = self.zorder)

		self.ax.set_xlabel("X (m)")
		self.ax.set_ylabel("Y (m)")

	def update(self):
		'''
		'''
		self.base_ax.set_data(self.grid.Z2D)
		if(self.hillshade_on):
			self.hillshade_ax.set_data(self.grid.hillshade)


	def drape_on(self, array, cmap = "gray", clim = None, delta_zorder = 1, alpha = 0.5, callback = None, callback_params = None):

		if clim is not None:
			vmin = clim[0]
			vmax = clim[1]
		else:
			vmin = np.nanmin(array)
			vmax = np.nanmax(array)


		ttaaxx = self.ax.imshow(array.reshape(self.grid.rshp), extent = self.grid.extent(), cmap = cmap, alpha = alpha, zorder = self.zorder + delta_zorder, vmin = vmin, vmax = vmax)

		if(callback == None):
			def nonecback():
				pass
			callback = nonecback


		return callbax_image(ttaaxx, callback, self.grid.rshp, clim = clim, callback_params = callback_params)

























































# dskjhksadfhla


'''
Steenbock's implementation of stream power models
Multiple styles, I'll start with functional

B.G - 07/2024 - Acigné

'''

import numba as nb
import numpy as np
from enum import Enum
import scabbard as scb


@nb.njit()
def _impl_spl_SFD_single(
		Stack, # TO order
		Sreceivers, # Steepest Receiver
		Sdx,   # distance to Srec
		Z,     # topography (edited in place)
		A,     # drainage area, or Discharge
		nx,    # Ncols
		ny,	   # Nrows
		dx,    # Spatial step
		BCs,   # Bouncary codes
		K,     # Erodability
		m,     # Area exponent 
		n,     # Slope exponent
		dt,    # Time Step
	):
	
	# Traversing landscapes from downstream to upstream
	for node in Stack:

		# Ignoring outnodes
		if node == Sreceivers[node] or scb.ste.can_out_flat(node,BCs) or scb.ste.is_active_flat(node,BCs) == False:
			continue

		dzdx = (Z[node] - Z[Sreceivers[node]])/Sdx[node]

		if(dzdx<=0):
			Z[node] = Z[Sreceivers[node]] + 1e-4
			dzdx = 1e-4/dx

		factor = K * dt * (A[node])**m / (Sdx[node]**n);

		ielevation = Z[node];
		irec_elevation = Z[Sreceivers[node]];
		elevation_k = ielevation;
		elevation_prev = Z[node] + 500;
		tolerance = 1e-5;

		while (abs(elevation_k - elevation_prev) > tolerance) :
			elevation_prev = elevation_k
			slope = max(elevation_k - irec_elevation, 1e-6)
			diff = (elevation_k - ielevation + factor * slope**n) / (1. + factor * n * slope**(n - 1))
			elevation_k -= diff;
		
		Z[node] = elevation_k

@nb.njit()
def _impl_spl_SFD_variable(
		Stack, # TO order
		Sreceivers, # Steepest Receiver
		Sdx,   # distance to Srec
		Z,     # topography (edited in place)
		A,     # drainage area, or Discharge
		nx,    # Ncols
		ny,	   # Nrows
		dx,    # Spatial step
		BCs,   # Bouncary codes
		K,     # Erodability
		m,     # Area exponent 
		n,     # Slope exponent
		dt,    # Time Step
	):
	
	# Traversing landscapes from downstream to upstream
	for node in Stack:

		# Ignoring ounodes
		if node == Sreceivers[node] or scb.ste.can_out_flat(node,BCs) or scb.ste.is_active_flat(node,BCs) == False:
			continue

		dzdx = (Z[node] - Z[Sreceivers[node]])/Sdx[node]

		if(dzdx<=0):
			Z[node] = Z[Sreceivers[node]] + 1e-4
			dzdx = 1e-4/dx

		factor = K[node] * dt * (A[node])**m / (Sdx[node]**n);

		ielevation = Z[node];
		irec_elevation = Z[Sreceivers[node]];
		elevation_k = ielevation;
		elevation_prev = Z[node] + 500;
		tolerance = 1e-5;

		while (abs(elevation_k - elevation_prev) > tolerance) :
			elevation_prev = elevation_k
			slope = max(elevation_k - irec_elevation, 1e-6)
			diff = (elevation_k - ielevation + factor * slope**n) / (1. + factor * n * slope**(n - 1))
			elevation_k -= diff;
		
		Z[node] = elevation_k

def run_SPL_on_topo(
		dem,     # topography (edited in place)
		BCs = None,   # Bouncary codes
		graph = None,
		K = 1e-5,     # Erodability
		m = 0.45,     # Area exponent 
		n = 1.11,     # Slope exponent
		dt = 1e3,    # Time Step	
	):
	dx,nx,ny = dem.geo.dxnxny

	if BCs is None:
		BCs = scb.ut.normal_BCs_from_shape(nx, ny, out_code = 3)

	if(graph is None):
		graph = scb.flow.SFGraph(dem, BCs = BCs, D4 = False, dx = 1., backend = 'ttb', fill_LM = True, step_fill = 1e-3)
	else:
		graph.update(dem, BCs = BCs, fill_LM = True, step_fill = 1e-3)

	A = scb.flow.drainage_area(graph)


	if(isinstance(K,np.ndarray)):
		_impl_spl_SFD_variable(graph.Stack, graph.Sreceivers, graph.Sdx, dem.Z.ravel(), A.ravel(), nx, ny, dx, BCs.ravel(), K.ravel(), m, n, dt)
	else:
		_impl_spl_SFD_single(graph.Stack, graph.Sreceivers, graph.Sdx, dem.Z.ravel(), A.ravel(), nx, ny, dx, BCs.ravel(), K, m, n, dt)
	




'''
Contains generic widgets customisation for matplotlib qt backend
For example specific types of sliders (float, min max)
'''

import sys
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_qtagg import FigureCanvas
from matplotlib.backends.backend_qtagg import NavigationToolbar2QT as NavigationToolbar

# from PySide6 import QtWidgets, QtCore
from matplotlib.backends.qt_compat import QtWidgets, QtCore
import matplotlib

# Ensure using the Qt6Agg backend with PySide6
matplotlib.use('QtAgg')






class FileLoader(QtWidgets.QWidget):

	theChosenFile = QtCore.Signal((str))
	
	def __init__(self):
		super(FileLoader, self).__init__()
		self.layout = QtWidgets.QVBoxLayout(self)

		# Button to open the file dialog
		self.openButton = QtWidgets.QPushButton("Open File")
		self.openButton.clicked.connect(self.openFileDialog)
		self.layout.addWidget(self.openButton)
		self.setMaximumSize(100,100)

	def openFileDialog(self):
		# Open a file dialog and print the selected file path
		filename, _ = QtWidgets.QFileDialog.getOpenFileName(self, "Open File", "", "All Files (*)")
		self.theChosenFile.emit(filename)
		if filename:
			print(f"Selected file: {filename}")


'''
grid module to help with generic grid manipulations
'''
import numpy as np
import matplotlib.pyplot as plt
from scabbard import legacy_load_raster
from scabbard import geography as geo
import dagger as dag
from scipy.ndimage import gaussian_filter
import random

class RGrid(object):

	"""
	Manages a regular grid with helper functions
	"""
	

	def __init__(self, nx, ny, dx, dy, Z, geography = None, dtype = np.float32):

		'''
		Contruct a grid from base char:
		- nx = number of cols
		- ny = number of rows
		- dx = spacing in the X direction
		- dy = spacing in the Y direction
		- Z = a numpy array of ny * nx coordinates containing elevation data

		And optional specs:
		- geography: custom geogrphic info

		It is recommended to use premade automatic functions (bellow) as much as you can: they would create grid from loading a DEM or with random noise or slope surface for example

		'''

		# Ignore
		super(RGrid, self).__init__()
		
		# Number of col
		self.nx = nx
		# Number of rows
		self.ny = ny
		#nnodes
		self.nxy = self.nx * self.ny
		# Spatial step in X dir
		self.dx = dx
		# Spatial step in Y dir
		self.dy = dy

		self.cellarea = dx*dy

		self.lx = (nx+1) * dx
		self.ly = (ny+1) * dy

		# Converts 1D flattened to 2D grid
		self.rshp = (ny,nx)

		if(geography is None):
			self.geography = geo.geog(xmin = 0., ymin = 0., xmax = self.lx, ymax = self.ly)
		else:
			self.geography = geography

		# Elevation flat array 
		self._Z = Z.ravel().astype(dtype)

		# Placeholders for connector and graph objects
		self.con = None
		self.graph = None

		

	def extent(self, y_min_top = True):
		'''
		Bounding box for the dem: [xmin,xmax, ymin, ymax].
		Can be directly used in matplotlib's imshow function
		'''
		return [self.geography.xmin, self.geography.xmax, self.geography.ymin if y_min_top else self.geography.ymax, self.geography.ymax if y_min_top else self.geography.ymin ]


	@property
	def X(self):
		'''
		1D array [nx] of flat X coordinates
		'''
		return np.linspace(self.geography.xmin + self.dx/2, self.geography.xmax - self.dx/2, self.nx)

	@property
	def Y(self):
		'''
		1D array [ny] of flat Y coordinates
		'''
		return np.linspace(self.geography.ymin + self.dy/2, self.geography.ymax - self.dy/2, self.ny)

	@property
	def XY(self):
		'''
		2D array of XY coordinates for each node -  as a  meshgrid.
		'''
		xx,yy = np.meshgrid(self.X, self.Y)
		return xx, yy[::-1]

	@property
	def Z(self):
		'''
		Accessing the 1D flat elevation array [nx * ny]
		'''
		return self._Z

	@property
	def Z2D(self):
		'''
		Accessing the 2D elevation array [ny,nx]
		'''
		return self._Z.reshape(self.ny,self.nx)

	@property
	def XYZ(self):
		'''
		meshgrid of XYZ coordinates
		'''
		xx,yy = np.meshgrid(self.X, self.Y)
		return xx, yy, self.Z2D

	@property
	def hillshade(self):
		'''
		2D array of hillshaded relief (values between 0 and 1)
		'''
		if(self.con is None):
			con = dag.D8N(self.nx, self.ny, self.dx, self.dy, self.geography.xmin, self.geography.ymin)
		else:
			con = self.con
		return dag.hillshade(con, self._Z).reshape(self.rshp)

	
	def rayshade(self,ray_slope = 0.75, shadow_mag = 1.,smooth_r = 2, attenuation = 0.95, radial_lights = True):
		if(self.con is None or self.graph is None):
			print("cannot rayshade is graph and con not assigned to the grid, if you have an external graphcon, directly call the dagger raysahde function")
		else:
			return dag.rayshade(self.graph, self.con, self._Z, ray_slope,shadow_mag,smooth_r, attenuation, radial_lights).reshape(self.rshp)

	def export_graphcon(self, process = True):
		
		con = dag.D8N(self.nx, self.ny, self.dx, self.dy, self.geography.xmin, self.geography.ymin)
		graph = dag.graph(con)
		if(process):
			graph.compute_graph(self._Z, True, False)
		return graph, con

	def compute_graphcon(self, SFD = False, BCs = None, LM = dag.LMR.none, preprocess_topo = False, Z0 = None ):
		self.con = dag.D8N(self.nx, self.ny, self.dx, self.dy, self.geography.xmin, self.geography.ymin)
		
		if(BCs is not None):
			self.con.set_custom_boundaries(BCs.ravel())

		if(Z0 is not None):
			dag.set_BC_to_remove_seas(self.con, self._Z, Z0) 
		
		self.graph = dag.graph(self.con)
		self.graph.set_LMR_method(LM)
		if(preprocess_topo):
			self._Z = self.graph.compute_graph(self._Z, SFD, False)
		else:
			self.graph.compute_graph(self._Z, SFD, False)

	def get_fill_topo(self, ret2D = False):
		'''
		Standalone topo filler using priority flood (Barnes, 2014)
		'''

		otopo = self._Z.copy()
		if(self.graph is None):
			self.compute_graphcon(SFD = True, BCs = None, LM = dag.LMR.priority_flood, preprocess_topo = True)
			ret = self._Z.copy()
		else:
			ret = dag.standalone_priority_flood(self._Z, self.connector)

		self._Z = otopo
		return ret if ret2D == False else ret.reshape(self.ny,self.nx)

	def zeros(self):
		return np.zeros(self.rshp)

	def min(self):
		return np.nanmin(self._Z)

	def max(self):
		return np.nanmax(self._Z)

	def add_random_noise(self, rmin = 0, rmax = 1):
		self._Z += np.random.uniform(low=rmin, high=rmax, size=(self.nxy,))

	def get_normalised(self, dim2 = True):
		'''
		'''
		return (self.Z2D - self.min())/(self.max() - self.min())

	def quick_river_network(self, DAT = 1e6, custom_QA = None):
		
		if(self.graph is None):
			self.compute_graphcon(SFD = True, BCs = None, LM = dag.LMR.cordonnier_carve, preprocess_topo = False)
		A = self.graph.accumulate_constant_downstream_SFD(self.cellarea) if (custom_QA is None) else custom_QA

		rivdict = dag.RiverNetwork(DAT, A,self._Z, self.con, self.graph)

		rivdict["rows"] = rivdict["nodes"] // self.nx
		rivdict["cols"] = rivdict["nodes"] % self.nx
		rivdict["X"] = self.XY[0].ravel()[rivdict["nodes"]]
		rivdict["Y"] = self.XY[1].ravel()[rivdict["nodes"]]

		return rivdict

	def quick_basin_extraction(self, basinID = None, return_basinID = False):
		if(self.graph is None):
			self.compute_graphcon(SFD = True, BCs = None, LM = dag.LMR.cordonnier_carve, preprocess_topo = False)

		if(basinID is None):
			basinID = self.graph.get_SFD_basin_labels()

		DDdict = dag.DrainageDivides(self.con, self.graph, self._Z, basinID.ravel())
		DDdict["rows"] = DDdict["nodes"] // self.nx
		DDdict["cols"] = DDdict["nodes"] % self.nx
		DDdict["X"] = self.XY[0].ravel()[DDdict["nodes"]]
		DDdict["Y"] = self.XY[1].ravel()[DDdict["nodes"]]


		if(return_basinID):
			return DDdict, basinID.reshape(self.rshp)
		else:
			return DDdict

	def baseplot(self, alpha_hs = 0.65, cmap = "gist_earth", clim = None, figsize = None, fig = None, ax = None, colorbar = True ):
		need_return = False
		if(fig is None):
			fig,ax = plt.subplots(figsize = figsize)
			need_return = True
		cb = ax.imshow(self.Z2D, cmap = cmap, vmin = self._Z.min() if clim is None else clim[0], vmax = self._Z.max() if clim is None else clim[1], extent = self.extent())
		
		if(colorbar):
			plt.colorbar(cb, label = "elevation (m)")

		ax.imshow(self.hillshade, cmap = "gray", vmin = 0, vmax = 1, alpha = alpha_hs, extent = self.extent())
		ax.set_xlabel("X (m)")
		ax.set_ylabel("Y (m)")
		
		if(need_return):
			return fig,ax





	def __str__(self):
		return f"""
Regular Grid object
nx = {self.nx} ny = {self.ny} -> n nodes = {self.nxy}
lx = {self.lx} ly = {self.ly} 
dx = {self.dx} dy = {self.dy}
"""








def generate_noise_RGrid( 
	nx = 256, ny = 256, dx = 30., dy = 30., # Dimensions
	noise_type = "white", # noise type: white or Perlin
	magnitude = 1,
	frequency = 4., octaves = 8, seed = None, # Perlin noise options and seed
	n_gaussian_smoothing = 0 # seed
	):
	
	
	if(noise_type.lower() == "white"):
		Z = np.random.rand(nx*ny) * magnitude
	elif(noise_type.lower() == "perlin"):
		con = dag.D8N(nx,ny,dx,dy,0,0)
		Z = dag.generate_perlin_noise_2D(con, frequency, octaves, np.uint32(seed) if seed is not None else np.uint32(random.randrange(0,32000) ) )

	if(n_gaussian_smoothing > 0):
		Z = gaussian_filter(Z,n_gaussian_smoothing)
	Z = Z.reshape(ny, nx)
	Z[[0,-1],:] = 0
	Z[:, [-1,0]] = 0

	return RGrid(nx, ny, dx, dy, Z, geography = None)



def raster2RGrid(fname, dtype = np.float32):

	dem = legacy_load_raster(fname)
	geog = geo.geog(dem["x_min"],dem["y_min"],dem["x_max"],dem["y_max"],dem["crs"])
	return RGrid(dem["nx"], dem["ny"], dem["dx"], dem["dy"], dem["array"].ravel().astype(dtype), geography=geog, dtype = dtype)



def slope_RGrid(
	nx = 512, 
	ny = 512, 
	dx = 5,
	dy = 5,
	z_base = 0,
	slope = 1e-3, 
	noise_magnitude = 1,
	EW = "periodic",
	S = "force",
	N = "force"
	):
	'''
		Generates a slopping grid. It comes with a connector and a graph set up so that the top part of the topo inputs fluxes and the bottom part output
	'''

	# Zeros Topo
	Z = np.zeros(ny * nx)

	# Generating the grid
	grid = RGrid(nx, ny, dx, dy, Z)

	# Getting the matrix
	X,Y,Z = grid.XYZ

	Z = - slope * Y
	Z += z_base -  Z.min()

	grid._Z = Z.ravel()

	if(noise_magnitude > 0):
		grid.add_random_noise(0, noise_magnitude)


	bc = np.ones((ny,nx),dtype = np.int32)
	bc[:,[0,-1]] = 9 if EW == "periodic" else (0 if EW == "noflow" else (4 if EW == "out" else 3)) #9 is periodic, 0 is no flow, 4 is normal out
	bc[0,:] = 8 if N == "force" else (0 if N == "noflow" else (4 if N == "out" else 3))
	bc[-1,:] = 5 if S == "force" else (4 if S == "out" else 3)
	
	if(EW == "noflow" or EW == "periodic"):
		bc[[0,-1],0] = 0
		bc[[0,-1],-1] = 0
	

	con = dag.D8N(nx,ny,dx,dy,0,0)
	con.set_custom_boundaries(bc.ravel())
	
	graph = dag.graph(con)


	graph.compute_graph(grid._Z, True, False)

	grid.con = con
	grid.graph = graph

	return grid



# // TO DO
# def gridFromTrackscape(ts):
# 	RGrid(nx, ny, dx, dy, Z, geography = None, dtype = np.float32)































































# end of file


'''
High-level interface to graphflood, starting point to run the model to other backend
'''

import numpy as np
import scabbard as scb
import dagger as dag
import taichi as ti

def _std_run_gpu_ndt(
	grid,
	P = None, # precipitations, numpy array or scalar
	BCs = None, # Boundary codes
	N_dt = 5000,
	dt = 1e-3,
	manning = 0.033,
	init_hw = None,

):
	'''Internal runner function for the standalone gpu runner (Runs N iteration from "scratch")'''
	try:
		ti.init(ti.gpu)
	except:
		raise RuntimeError('Could not initiate taichi to GPU')

	# Creating the parameter file
	param = scb.rvd.param_from_grid(grid)
	param.manning = manning

	# Feeding in initial hw
	if(init_hw is not None):
		param.initial_hw = init_hw

	# Feeding the BCs (if not default to flow everywhere, and can out from the edges)
	if(BCs is not None):
		param.BCs = BCs

	# Simulation time steps
	param.dt_hydro = dt

	# Field of precipitation inputs
	param.precipitations = P

	# COmpiling param to model object
	rd = scb.rvd.create_from_params(param)

	# Running the N time steps
	rd.run_hydro(N_dt)

	# Return metrics
	return {'h':rd.hw.to_numpy(), 'Qi': rd.QwA.to_numpy(), 'Qo':rd.QwC.to_numpy(), 'model':rd, 'param':param, 'backend_graphflood': 'riverdale'}


def _std_rerun_gpu_ndt(
	model_input,
	N_dt = 1000,
	new_dt = None):

	if(new_dt is not None):
		scb.rvd.change_dt_hydro(model_input['model'], model_input['param'], new_dt)

	rd = model_input['model']
	# Running the N time steps
	rd.run_hydro(N_dt)

	# Return metrics
	return {'h':rd.hw.to_numpy(), 'Qi': rd.QwA.to_numpy(), 'Qo':rd.QwC.to_numpy(), 'model':rd, 'param':param, 'backend_graphflood': 'riverdale'}

def _std_run_dagger_ndt(
	grid,
	P = None, # precipitations, numpy array or scalar
	BCs = None,
	N_dt = 5000,
	dt = 1e-3,
	manning = 0.033,
	init_hw = None,

	**kwargs

	):
	'''
		Internal function managing the DAGGER interface
	'''

	# init the DAGGER engine (connector and graph fix the topology of the grid)
	con = dag.D8N(grid.geo.nx, grid.geo.ny, grid.geo.dx, grid.geo.dx, grid.geo.xmin, grid.geo.ymin)
	graph = dag.graph(con)

	# Eventual custom boundary conditions
	if(BCs is not None):
		con.set_custom_boundaries(BCs.ravel())

	# Initialising the c++ graphflood object
	flood = dag.graphflood(graph, con)

	# feeding the topo
	flood.set_topo(grid.Z.ravel())
	
	# Setting the topology
	setsfd = False 
	if ('SFD' in kwargs.keys()):
		if(kwargs['SFD']):
			setsfd = True
	flood.enable_SFD() if setsfd else flood.enable_MFD() 

	# Legacy options, ignore
	flood.fill_minima()
	flood.disable_courant_dt_hydro()

	# Setting the time step
	flood.set_dt_hydro(dt)

	flood.set_mannings(manning)

	if(isinstance(P, np.ndarray)):
		flood.set_water_input_by_variable_precipitation_rate(P.ravel())
	else:
		flood.set_water_input_by_constant_precipitation_rate(P)

	if(init_hw is not None):
		flood.set_hw(init_hw.ravel())

	for i in range(N_dt):
		flood.run()

	return {'h':flood.get_hw().reshape(grid.rshp), 'Qi': flood.get_Qwin().reshape(grid.rshp), 
		'Qo':flood.compute_tuqQ(3).reshape(grid.rshp), 'model':flood, 'param':None, 'backend_graphflood': 'dagger'}


def _std_run_ttb_ndt(
	grid,
	P = None, # precipitations, numpy array or scalar
	BCs = None,
	N_dt = 5000,
	dt = 1e-3,
	manning = 0.033,
	init_hw = None,

	**kwargs
	):
	
	import topotoolbox as ttb

	ttbgrid = grid.grid2ttb()

	sfd = False
	if('SFD' in kwargs.keys()):
		if(kwargs['SFD'] == True):
			sfd = True

	D8 = True
	if('D4' in kwargs.keys()):
		if(kwargs['D4'] == True):
			D8 = False
	
	res = ttb.run_graphflood(
		ttbgrid,
		initial_hw=init_hw,
		bcs=BCs,
		dt=dt,
		p=P,
		manning=manning,
		sfd=sfd,
		d8=D8,
		n_iterations=N_dt
	)

	return {'h': scb.raster.raster_from_ttb(res),'Qi': None,'Qo':None, 'model':None, 'param':None, 'backend_graphflood': 'ttb'}

def std_run(
	model_input, # Sting or grid so far
	P = 1e-5, # precipitations, numpy array or scalar
	BCs = None, # Boundary codes
	N_dt = 5000,
	backend = 'gpu',
	dt = 1e-3,
	manning = 0.033,
	init_hw = None,
	**kwargs
):

	if(isinstance(model_input, str)):
		grid = scb.io.load_raster(model_input)
	else:
		grid = model_input
	
	if(backend.lower() == 'gpu'):
		return _std_run_gpu_ndt(grid, P, BCs, N_dt, dt, manning, init_hw = init_hw, **kwargs)
	elif(backend.lower() in [ 'dagger' , 'cpu']):
		return _std_run_dagger_ndt(grid, P,	BCs, N_dt, dt, manning, init_hw, **kwargs)
	elif(backend.lower() in ['ttb', 'topotoolbox']):
		return _std_run_ttb_ndt(grid, P, BCs, N_dt, dt, manning, init_hw, **kwargs)



def std_rerun(model_input, N_dt = 1000, new_dt = None):
	'''function to run further dts from an existing model output
	'''

	# Checking if I can run these

	if(isinstance(model_input,dict) == False):
		raise RuntimeError(f'std_rerun takes an existing ouput from std_run to run it further, not a {type(model_input)}.')
	elif('backend_graphflood' not in model_input):
		raise RuntimeError(f'std_rerun takes an existing ouput from std_run to run it further, not a {type(model_input)}.')

	# Else starting again:



"""
This class deals with loading raster informations
Authors: B.G.
"""
import numpy as np
import rasterio as rio
from rasterio.transform import from_bounds
from rasterio.transform import from_origin
import scabbard as scb

def load_raster(fname, dtype = np.float32):
	"""
	Load a raster file into a RegularRasterGrid object
	Made to read DEMs and DEM-like data (i.e. single band geolocalised raster data), does not handle multi-band.
	I can make a multi-band option if needed
	Uses rasterio and gdal behind the scene, see https://gdal.org/drivers/raster/index.html for accepted formats

	Arguments:
		- fname (str or any sort of path): full file name + path if not local
		- dtype (numpy type): forces a type to the raster, by default float 32 bits (single recision) 
	
	Returns:
		- The raster object with data loaded
	
	Authors:
		- B.G. (last modification: 08/2024)

	"""

	# Loading the raster with rasterio
	this_raster = rio.open(fname)

	# Getting the resolution
	gt = this_raster.res
	
	# Creating the underlying geomoetry object
	geom = scb.geometry.RegularGeometry(this_raster.width, this_raster.height, gt[0], this_raster.bounds[0], this_raster.bounds[1])
	
	# Getting the actual data
	Z = this_raster.read(1).astype(dtype)
	
	# NO no data handling so far
	# Z[Z == this_raster.nodatavals] = np.nan

	# Checks if the DEM has a crs or not
	try:
		geom._crs = this_raster.crs['init']
	except (TypeError, KeyError) as e:
		geom._crs = u'epsg:32601'

	return scb.raster.RegularRasterGrid(Z, geom, dtype = dtype)